"""
Enterprise Throttle Module.

Provides request throttling, rate control, backpressure,
and adaptive load management.

Example:
    # Create throttle
    throttle = create_throttle(rate=100, period=60)  # 100 requests per minute
    
    # Check if allowed
    if await throttle.acquire():
        # Process request
        ...
    else:
        # Rate limited
        ...
    
    # With decorator
    @throttled(rate=10, period=1)  # 10 per second
    async def api_call():
        ...
    
    # Adaptive throttle
    adaptive = create_adaptive_throttle(
        min_rate=10,
        max_rate=1000,
        target_latency_ms=100
    )
"""

from __future__ import annotations

import asyncio
import logging
import time
from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from functools import wraps
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
)

logger = logging.getLogger(__name__)

T = TypeVar('T')


class ThrottleError(Exception):
    """Throttle error."""
    pass


class RateLimitExceeded(ThrottleError):
    """Rate limit exceeded."""
    pass


class ThrottleStrategy(str, Enum):
    """Throttle strategies."""
    FIXED_WINDOW = "fixed_window"
    SLIDING_WINDOW = "sliding_window"
    TOKEN_BUCKET = "token_bucket"
    LEAKY_BUCKET = "leaky_bucket"
    ADAPTIVE = "adaptive"


class ThrottleAction(str, Enum):
    """Action when throttled."""
    REJECT = "reject"
    QUEUE = "queue"
    DELAY = "delay"


@dataclass
class ThrottleConfig:
    """Throttle configuration."""
    rate: int  # Requests per period
    period: float = 1.0  # Period in seconds
    burst: int = 0  # Extra burst capacity
    strategy: ThrottleStrategy = ThrottleStrategy.SLIDING_WINDOW
    action: ThrottleAction = ThrottleAction.REJECT
    queue_size: int = 100


@dataclass
class ThrottleStats:
    """Throttle statistics."""
    total_requests: int = 0
    allowed_requests: int = 0
    throttled_requests: int = 0
    current_rate: float = 0.0
    avg_wait_time_ms: float = 0.0


@dataclass
class ThrottleResult:
    """Result of throttle check."""
    allowed: bool
    wait_time: float = 0.0
    remaining: int = 0
    reset_at: Optional[datetime] = None
    retry_after: Optional[float] = None


class Throttle(ABC):
    """Abstract throttle interface."""
    
    @abstractmethod
    async def acquire(self, count: int = 1) -> ThrottleResult:
        """Try to acquire tokens."""
        pass
    
    @abstractmethod
    async def release(self, count: int = 1) -> None:
        """Release tokens (if applicable)."""
        pass
    
    @abstractmethod
    def get_stats(self) -> ThrottleStats:
        """Get throttle statistics."""
        pass
    
    @abstractmethod
    def reset(self) -> None:
        """Reset the throttle."""
        pass


class FixedWindowThrottle(Throttle):
    """Fixed window rate limiting."""
    
    def __init__(self, config: ThrottleConfig):
        self._config = config
        self._window_start = time.time()
        self._count = 0
        self._stats = ThrottleStats()
    
    async def acquire(self, count: int = 1) -> ThrottleResult:
        """Check if request is allowed."""
        now = time.time()
        self._stats.total_requests += 1
        
        # Check if window has expired
        if now - self._window_start >= self._config.period:
            self._window_start = now
            self._count = 0
        
        # Check rate limit
        limit = self._config.rate + self._config.burst
        
        if self._count + count <= limit:
            self._count += count
            self._stats.allowed_requests += 1
            
            return ThrottleResult(
                allowed=True,
                remaining=limit - self._count,
                reset_at=datetime.fromtimestamp(self._window_start + self._config.period),
            )
        
        self._stats.throttled_requests += 1
        wait_time = self._config.period - (now - self._window_start)
        
        return ThrottleResult(
            allowed=False,
            remaining=0,
            reset_at=datetime.fromtimestamp(self._window_start + self._config.period),
            retry_after=wait_time,
            wait_time=wait_time,
        )
    
    async def release(self, count: int = 1) -> None:
        """Release is no-op for fixed window."""
        pass
    
    def get_stats(self) -> ThrottleStats:
        """Get statistics."""
        return self._stats
    
    def reset(self) -> None:
        """Reset the throttle."""
        self._window_start = time.time()
        self._count = 0


class SlidingWindowThrottle(Throttle):
    """Sliding window rate limiting."""
    
    def __init__(self, config: ThrottleConfig):
        self._config = config
        self._requests: Deque[float] = deque()
        self._stats = ThrottleStats()
    
    async def acquire(self, count: int = 1) -> ThrottleResult:
        """Check if request is allowed."""
        now = time.time()
        self._stats.total_requests += 1
        
        # Remove expired entries
        cutoff = now - self._config.period
        while self._requests and self._requests[0] < cutoff:
            self._requests.popleft()
        
        # Check rate limit
        limit = self._config.rate + self._config.burst
        
        if len(self._requests) + count <= limit:
            for _ in range(count):
                self._requests.append(now)
            
            self._stats.allowed_requests += 1
            
            return ThrottleResult(
                allowed=True,
                remaining=limit - len(self._requests),
            )
        
        self._stats.throttled_requests += 1
        
        # Calculate wait time
        wait_time = 0.0
        if self._requests:
            oldest = self._requests[0]
            wait_time = max(0, self._config.period - (now - oldest))
        
        return ThrottleResult(
            allowed=False,
            remaining=0,
            retry_after=wait_time,
            wait_time=wait_time,
        )
    
    async def release(self, count: int = 1) -> None:
        """Release is no-op for sliding window."""
        pass
    
    def get_stats(self) -> ThrottleStats:
        """Get statistics."""
        self._stats.current_rate = len(self._requests) / self._config.period
        return self._stats
    
    def reset(self) -> None:
        """Reset the throttle."""
        self._requests.clear()


class TokenBucketThrottle(Throttle):
    """Token bucket rate limiting."""
    
    def __init__(self, config: ThrottleConfig):
        self._config = config
        self._tokens = float(config.rate + config.burst)
        self._max_tokens = float(config.rate + config.burst)
        self._refill_rate = config.rate / config.period
        self._last_refill = time.time()
        self._stats = ThrottleStats()
        self._lock = asyncio.Lock()
    
    async def acquire(self, count: int = 1) -> ThrottleResult:
        """Try to acquire tokens."""
        async with self._lock:
            self._stats.total_requests += 1
            self._refill()
            
            if self._tokens >= count:
                self._tokens -= count
                self._stats.allowed_requests += 1
                
                return ThrottleResult(
                    allowed=True,
                    remaining=int(self._tokens),
                )
            
            self._stats.throttled_requests += 1
            
            # Calculate wait time for tokens
            needed = count - self._tokens
            wait_time = needed / self._refill_rate
            
            return ThrottleResult(
                allowed=False,
                remaining=0,
                retry_after=wait_time,
                wait_time=wait_time,
            )
    
    async def release(self, count: int = 1) -> None:
        """Return tokens to bucket."""
        async with self._lock:
            self._tokens = min(self._max_tokens, self._tokens + count)
    
    def _refill(self) -> None:
        """Refill tokens based on elapsed time."""
        now = time.time()
        elapsed = now - self._last_refill
        self._last_refill = now
        
        refill = elapsed * self._refill_rate
        self._tokens = min(self._max_tokens, self._tokens + refill)
    
    def get_stats(self) -> ThrottleStats:
        """Get statistics."""
        self._stats.current_rate = self._max_tokens - self._tokens
        return self._stats
    
    def reset(self) -> None:
        """Reset the throttle."""
        self._tokens = self._max_tokens
        self._last_refill = time.time()


class LeakyBucketThrottle(Throttle):
    """Leaky bucket rate limiting (smooth output rate)."""
    
    def __init__(self, config: ThrottleConfig):
        self._config = config
        self._queue: Deque[Tuple[float, asyncio.Event]] = deque()
        self._leak_rate = config.rate / config.period  # requests per second
        self._last_leak = time.time()
        self._stats = ThrottleStats()
        self._running = False
        self._task: Optional[asyncio.Task] = None
    
    async def start(self) -> None:
        """Start the leaky bucket processor."""
        if self._running:
            return
        
        self._running = True
        self._task = asyncio.create_task(self._leak_loop())
    
    async def stop(self) -> None:
        """Stop the leaky bucket processor."""
        self._running = False
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
    
    async def _leak_loop(self) -> None:
        """Process queue at fixed rate."""
        interval = 1.0 / self._leak_rate
        
        while self._running:
            try:
                if self._queue:
                    _, event = self._queue.popleft()
                    event.set()
                
                await asyncio.sleep(interval)
                
            except asyncio.CancelledError:
                break
    
    async def acquire(self, count: int = 1) -> ThrottleResult:
        """Queue request for processing."""
        self._stats.total_requests += 1
        
        # Check queue size
        if len(self._queue) >= self._config.queue_size:
            self._stats.throttled_requests += 1
            return ThrottleResult(
                allowed=False,
                remaining=0,
                wait_time=len(self._queue) / self._leak_rate,
            )
        
        # Add to queue and wait
        event = asyncio.Event()
        self._queue.append((time.time(), event))
        
        if self._config.action == ThrottleAction.QUEUE:
            await event.wait()
            self._stats.allowed_requests += 1
            return ThrottleResult(allowed=True, remaining=self._config.queue_size - len(self._queue))
        
        self._stats.allowed_requests += 1
        return ThrottleResult(
            allowed=True,
            remaining=self._config.queue_size - len(self._queue),
            wait_time=len(self._queue) / self._leak_rate,
        )
    
    async def release(self, count: int = 1) -> None:
        """Release is no-op for leaky bucket."""
        pass
    
    def get_stats(self) -> ThrottleStats:
        """Get statistics."""
        self._stats.current_rate = self._leak_rate
        return self._stats
    
    def reset(self) -> None:
        """Reset the throttle."""
        self._queue.clear()


class AdaptiveThrottle(Throttle):
    """
    Adaptive throttle that adjusts rate based on latency.
    """
    
    def __init__(
        self,
        min_rate: int = 10,
        max_rate: int = 1000,
        target_latency_ms: float = 100.0,
        adjustment_factor: float = 0.1,
    ):
        self._min_rate = min_rate
        self._max_rate = max_rate
        self._current_rate = (min_rate + max_rate) // 2
        self._target_latency = target_latency_ms
        self._adjustment_factor = adjustment_factor
        
        self._latencies: Deque[float] = deque(maxlen=100)
        self._inner_throttle = TokenBucketThrottle(ThrottleConfig(
            rate=self._current_rate,
            period=1.0,
        ))
        self._stats = ThrottleStats()
    
    async def acquire(self, count: int = 1) -> ThrottleResult:
        """Acquire with adaptive rate adjustment."""
        result = await self._inner_throttle.acquire(count)
        self._stats.total_requests += 1
        
        if result.allowed:
            self._stats.allowed_requests += 1
        else:
            self._stats.throttled_requests += 1
        
        return result
    
    async def release(self, count: int = 1) -> None:
        """Release tokens."""
        await self._inner_throttle.release(count)
    
    def record_latency(self, latency_ms: float) -> None:
        """Record a latency measurement and adjust rate."""
        self._latencies.append(latency_ms)
        
        if len(self._latencies) < 10:
            return
        
        avg_latency = sum(self._latencies) / len(self._latencies)
        
        # Adjust rate based on latency
        if avg_latency > self._target_latency:
            # Too slow, reduce rate
            adjustment = int(self._current_rate * self._adjustment_factor)
            self._current_rate = max(self._min_rate, self._current_rate - adjustment)
        else:
            # Fast enough, increase rate
            adjustment = int(self._current_rate * self._adjustment_factor)
            self._current_rate = min(self._max_rate, self._current_rate + adjustment)
        
        # Update inner throttle
        self._inner_throttle = TokenBucketThrottle(ThrottleConfig(
            rate=self._current_rate,
            period=1.0,
        ))
        
        logger.debug(f"Adjusted rate to {self._current_rate} (avg latency: {avg_latency:.1f}ms)")
    
    def get_stats(self) -> ThrottleStats:
        """Get statistics."""
        self._stats.current_rate = float(self._current_rate)
        
        if self._latencies:
            self._stats.avg_wait_time_ms = sum(self._latencies) / len(self._latencies)
        
        return self._stats
    
    def reset(self) -> None:
        """Reset the throttle."""
        self._latencies.clear()
        self._current_rate = (self._min_rate + self._max_rate) // 2
        self._inner_throttle.reset()


class ConcurrencyLimiter:
    """Limit concurrent operations."""
    
    def __init__(self, max_concurrent: int):
        self._max_concurrent = max_concurrent
        self._semaphore = asyncio.Semaphore(max_concurrent)
        self._current = 0
        self._stats = ThrottleStats()
    
    async def acquire(self) -> bool:
        """Acquire a slot."""
        self._stats.total_requests += 1
        
        acquired = await asyncio.wait_for(
            self._semaphore.acquire(),
            timeout=None,
        )
        
        if acquired:
            self._current += 1
            self._stats.allowed_requests += 1
        
        return acquired
    
    async def release(self) -> None:
        """Release a slot."""
        self._semaphore.release()
        self._current = max(0, self._current - 1)
    
    @property
    def current(self) -> int:
        """Get current concurrent count."""
        return self._current
    
    @property
    def available(self) -> int:
        """Get available slots."""
        return self._max_concurrent - self._current
    
    def get_stats(self) -> ThrottleStats:
        """Get statistics."""
        self._stats.current_rate = float(self._current)
        return self._stats


class BackpressureController:
    """
    Backpressure controller for flow control.
    """
    
    def __init__(
        self,
        low_watermark: int = 100,
        high_watermark: int = 1000,
    ):
        self._low_watermark = low_watermark
        self._high_watermark = high_watermark
        self._queue_size = 0
        self._paused = False
        self._on_pause: Optional[Callable] = None
        self._on_resume: Optional[Callable] = None
    
    @property
    def is_paused(self) -> bool:
        """Check if backpressure is active."""
        return self._paused
    
    def on_pause(self, callback: Callable) -> None:
        """Set pause callback."""
        self._on_pause = callback
    
    def on_resume(self, callback: Callable) -> None:
        """Set resume callback."""
        self._on_resume = callback
    
    def update(self, queue_size: int) -> None:
        """Update queue size and check thresholds."""
        self._queue_size = queue_size
        
        if not self._paused and queue_size >= self._high_watermark:
            self._paused = True
            logger.warning(f"Backpressure activated (queue size: {queue_size})")
            
            if self._on_pause:
                self._on_pause()
        
        elif self._paused and queue_size <= self._low_watermark:
            self._paused = False
            logger.info(f"Backpressure released (queue size: {queue_size})")
            
            if self._on_resume:
                self._on_resume()
    
    def check(self) -> bool:
        """Check if processing should continue."""
        return not self._paused


# Decorators
def throttled(
    rate: int,
    period: float = 1.0,
    strategy: ThrottleStrategy = ThrottleStrategy.SLIDING_WINDOW,
    on_throttled: Optional[Callable] = None,
) -> Callable:
    """
    Decorator to throttle function calls.
    
    Example:
        @throttled(rate=10, period=1)
        async def api_call():
            ...
    """
    config = ThrottleConfig(rate=rate, period=period, strategy=strategy)
    
    if strategy == ThrottleStrategy.TOKEN_BUCKET:
        throttle = TokenBucketThrottle(config)
    elif strategy == ThrottleStrategy.FIXED_WINDOW:
        throttle = FixedWindowThrottle(config)
    else:
        throttle = SlidingWindowThrottle(config)
    
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            result = await throttle.acquire()
            
            if not result.allowed:
                if on_throttled:
                    return on_throttled(result)
                raise RateLimitExceeded(f"Rate limit exceeded. Retry after {result.retry_after:.2f}s")
            
            return await func(*args, **kwargs)
        
        return wrapper
    
    return decorator


def concurrent_limit(max_concurrent: int) -> Callable:
    """
    Decorator to limit concurrent executions.
    
    Example:
        @concurrent_limit(5)
        async def process_item(item):
            ...
    """
    limiter = ConcurrencyLimiter(max_concurrent)
    
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            await limiter.acquire()
            try:
                return await func(*args, **kwargs)
            finally:
                await limiter.release()
        
        return wrapper
    
    return decorator


# Factory functions
def create_throttle(
    rate: int,
    period: float = 1.0,
    strategy: ThrottleStrategy = ThrottleStrategy.SLIDING_WINDOW,
    **kwargs: Any,
) -> Throttle:
    """Create a throttle."""
    config = ThrottleConfig(rate=rate, period=period, strategy=strategy, **kwargs)
    
    if strategy == ThrottleStrategy.FIXED_WINDOW:
        return FixedWindowThrottle(config)
    elif strategy == ThrottleStrategy.TOKEN_BUCKET:
        return TokenBucketThrottle(config)
    elif strategy == ThrottleStrategy.LEAKY_BUCKET:
        return LeakyBucketThrottle(config)
    else:
        return SlidingWindowThrottle(config)


def create_adaptive_throttle(
    min_rate: int = 10,
    max_rate: int = 1000,
    target_latency_ms: float = 100.0,
    **kwargs: Any,
) -> AdaptiveThrottle:
    """Create an adaptive throttle."""
    return AdaptiveThrottle(min_rate, max_rate, target_latency_ms, **kwargs)


def create_concurrency_limiter(max_concurrent: int) -> ConcurrencyLimiter:
    """Create a concurrency limiter."""
    return ConcurrencyLimiter(max_concurrent)


def create_backpressure_controller(
    low_watermark: int = 100,
    high_watermark: int = 1000,
) -> BackpressureController:
    """Create a backpressure controller."""
    return BackpressureController(low_watermark, high_watermark)


__all__ = [
    # Exceptions
    "ThrottleError",
    "RateLimitExceeded",
    # Enums
    "ThrottleStrategy",
    "ThrottleAction",
    # Data classes
    "ThrottleConfig",
    "ThrottleStats",
    "ThrottleResult",
    # Core classes
    "Throttle",
    "FixedWindowThrottle",
    "SlidingWindowThrottle",
    "TokenBucketThrottle",
    "LeakyBucketThrottle",
    "AdaptiveThrottle",
    "ConcurrencyLimiter",
    "BackpressureController",
    # Decorators
    "throttled",
    "concurrent_limit",
    # Factory
    "create_throttle",
    "create_adaptive_throttle",
    "create_concurrency_limiter",
    "create_backpressure_controller",
]



"""
Tool State Management.

Provides state tracking for tool execution, caching, and retries.
"""

import hashlib
import logging
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Callable, Dict, List, Optional

from .manager import StateManager, StateType

logger = logging.getLogger(__name__)


class ToolExecutionStatus(Enum):
    """Status of tool execution."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    TIMEOUT = "timeout"
    RETRYING = "retrying"
    CANCELLED = "cancelled"


@dataclass
class ToolExecution:
    """Single tool execution record."""
    execution_id: str
    tool_name: str
    status: ToolExecutionStatus
    args: Dict[str, Any]
    result: Any = None
    error: Optional[str] = None
    started_at: str = field(default_factory=lambda: datetime.now().isoformat())
    completed_at: Optional[str] = None
    duration_ms: int = 0
    retry_count: int = 0
    agent_id: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "execution_id": self.execution_id,
            "tool_name": self.tool_name,
            "status": self.status.value,
            "args": self.args,
            "result": self.result,
            "error": self.error,
            "started_at": self.started_at,
            "completed_at": self.completed_at,
            "duration_ms": self.duration_ms,
            "retry_count": self.retry_count,
            "agent_id": self.agent_id,
            "metadata": self.metadata,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ToolExecution":
        return cls(
            execution_id=data["execution_id"],
            tool_name=data["tool_name"],
            status=ToolExecutionStatus(data["status"]),
            args=data.get("args", {}),
            result=data.get("result"),
            error=data.get("error"),
            started_at=data.get("started_at", datetime.now().isoformat()),
            completed_at=data.get("completed_at"),
            duration_ms=data.get("duration_ms", 0),
            retry_count=data.get("retry_count", 0),
            agent_id=data.get("agent_id"),
            metadata=data.get("metadata", {}),
        )


@dataclass
class ToolCacheEntry:
    """Cached tool result."""
    cache_key: str
    tool_name: str
    args_hash: str
    result: Any
    cached_at: str = field(default_factory=lambda: datetime.now().isoformat())
    expires_at: Optional[str] = None
    hit_count: int = 0
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "cache_key": self.cache_key,
            "tool_name": self.tool_name,
            "args_hash": self.args_hash,
            "result": self.result,
            "cached_at": self.cached_at,
            "expires_at": self.expires_at,
            "hit_count": self.hit_count,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ToolCacheEntry":
        return cls(
            cache_key=data["cache_key"],
            tool_name=data["tool_name"],
            args_hash=data["args_hash"],
            result=data.get("result"),
            cached_at=data.get("cached_at", datetime.now().isoformat()),
            expires_at=data.get("expires_at"),
            hit_count=data.get("hit_count", 0),
        )
    
    @property
    def is_expired(self) -> bool:
        if not self.expires_at:
            return False
        return datetime.fromisoformat(self.expires_at) < datetime.now()


@dataclass
class RetryState:
    """Retry state for failed executions."""
    execution_id: str
    tool_name: str
    attempt: int
    max_attempts: int
    next_retry_at: str
    backoff_seconds: float
    last_error: Optional[str] = None
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "execution_id": self.execution_id,
            "tool_name": self.tool_name,
            "attempt": self.attempt,
            "max_attempts": self.max_attempts,
            "next_retry_at": self.next_retry_at,
            "backoff_seconds": self.backoff_seconds,
            "last_error": self.last_error,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "RetryState":
        return cls(
            execution_id=data["execution_id"],
            tool_name=data["tool_name"],
            attempt=data.get("attempt", 0),
            max_attempts=data.get("max_attempts", 3),
            next_retry_at=data.get("next_retry_at", datetime.now().isoformat()),
            backoff_seconds=data.get("backoff_seconds", 1.0),
            last_error=data.get("last_error"),
        )
    
    @property
    def can_retry(self) -> bool:
        return self.attempt < self.max_attempts
    
    @property
    def is_ready(self) -> bool:
        return datetime.fromisoformat(self.next_retry_at) <= datetime.now()


@dataclass
class ToolStats:
    """Statistics for a tool."""
    tool_name: str
    total_executions: int = 0
    successful_executions: int = 0
    failed_executions: int = 0
    total_duration_ms: int = 0
    avg_duration_ms: float = 0.0
    cache_hits: int = 0
    cache_misses: int = 0
    last_executed: Optional[str] = None
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "tool_name": self.tool_name,
            "total_executions": self.total_executions,
            "successful_executions": self.successful_executions,
            "failed_executions": self.failed_executions,
            "total_duration_ms": self.total_duration_ms,
            "avg_duration_ms": self.avg_duration_ms,
            "cache_hits": self.cache_hits,
            "cache_misses": self.cache_misses,
            "last_executed": self.last_executed,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ToolStats":
        return cls(
            tool_name=data["tool_name"],
            total_executions=data.get("total_executions", 0),
            successful_executions=data.get("successful_executions", 0),
            failed_executions=data.get("failed_executions", 0),
            total_duration_ms=data.get("total_duration_ms", 0),
            avg_duration_ms=data.get("avg_duration_ms", 0.0),
            cache_hits=data.get("cache_hits", 0),
            cache_misses=data.get("cache_misses", 0),
            last_executed=data.get("last_executed"),
        )


class ToolStateManager:
    """
    Manages state for tool execution, caching, and retries.
    
    Example:
        >>> tool_state = ToolStateManager()
        >>> 
        >>> # Start execution
        >>> exec = tool_state.start_execution("search", {"query": "AI"})
        >>> 
        >>> # Complete execution
        >>> tool_state.complete_execution(exec.execution_id, result={"items": [...]})
        >>> 
        >>> # Use caching
        >>> cached = tool_state.get_cached_result("search", {"query": "AI"})
        >>> if not cached:
        ...     result = run_search(...)
        ...     tool_state.cache_result("search", {"query": "AI"}, result)
    """
    
    def __init__(
        self,
        state_manager: StateManager = None,
        default_cache_ttl: int = 3600,  # 1 hour
        default_max_retries: int = 3,
        default_backoff_base: float = 2.0,
    ):
        self.state_manager = state_manager or StateManager()
        self.default_cache_ttl = default_cache_ttl
        self.default_max_retries = default_max_retries
        self.default_backoff_base = default_backoff_base
    
    def _hash_args(self, args: Dict[str, Any]) -> str:
        """Create hash from arguments."""
        import json
        args_str = json.dumps(args, sort_keys=True, default=str)
        return hashlib.sha256(args_str.encode()).hexdigest()[:16]
    
    # Execution Tracking
    def start_execution(
        self,
        tool_name: str,
        args: Dict[str, Any],
        agent_id: str = None,
        metadata: Dict = None,
    ) -> ToolExecution:
        """Start a tool execution."""
        import uuid
        execution_id = f"exec-{uuid.uuid4().hex[:8]}"
        
        execution = ToolExecution(
            execution_id=execution_id,
            tool_name=tool_name,
            status=ToolExecutionStatus.RUNNING,
            args=args,
            agent_id=agent_id,
            metadata=metadata or {},
        )
        
        self.state_manager.save(
            f"tool_exec:{execution_id}",
            execution.to_dict(),
            StateType.TOOL,
        )
        
        return execution
    
    def complete_execution(
        self,
        execution_id: str,
        result: Any,
        cache: bool = False,
        cache_ttl: int = None,
    ) -> bool:
        """Mark execution as completed."""
        data = self.state_manager.get(f"tool_exec:{execution_id}")
        if not data:
            return False
        
        execution = ToolExecution.from_dict(data)
        execution.status = ToolExecutionStatus.COMPLETED
        execution.result = result
        execution.completed_at = datetime.now().isoformat()
        
        # Calculate duration
        started = datetime.fromisoformat(execution.started_at)
        duration = (datetime.now() - started).total_seconds() * 1000
        execution.duration_ms = int(duration)
        
        self.state_manager.save(
            f"tool_exec:{execution_id}",
            execution.to_dict(),
            StateType.TOOL,
        )
        
        # Update stats
        self._update_stats(execution.tool_name, success=True, duration_ms=execution.duration_ms)
        
        # Cache result if requested
        if cache:
            self.cache_result(
                execution.tool_name,
                execution.args,
                result,
                ttl=cache_ttl or self.default_cache_ttl,
            )
        
        return True
    
    def fail_execution(
        self,
        execution_id: str,
        error: str,
        schedule_retry: bool = False,
    ) -> Optional[RetryState]:
        """Mark execution as failed."""
        data = self.state_manager.get(f"tool_exec:{execution_id}")
        if not data:
            return None
        
        execution = ToolExecution.from_dict(data)
        execution.status = ToolExecutionStatus.FAILED
        execution.error = error
        execution.completed_at = datetime.now().isoformat()
        
        # Calculate duration
        started = datetime.fromisoformat(execution.started_at)
        duration = (datetime.now() - started).total_seconds() * 1000
        execution.duration_ms = int(duration)
        
        self.state_manager.save(
            f"tool_exec:{execution_id}",
            execution.to_dict(),
            StateType.TOOL,
        )
        
        # Update stats
        self._update_stats(execution.tool_name, success=False, duration_ms=execution.duration_ms)
        
        # Schedule retry if requested
        if schedule_retry:
            return self._schedule_retry(execution, error)
        
        return None
    
    def cancel_execution(self, execution_id: str) -> bool:
        """Cancel an execution."""
        data = self.state_manager.get(f"tool_exec:{execution_id}")
        if not data:
            return False
        
        execution = ToolExecution.from_dict(data)
        if execution.status in [ToolExecutionStatus.COMPLETED, ToolExecutionStatus.FAILED]:
            return False
        
        execution.status = ToolExecutionStatus.CANCELLED
        execution.completed_at = datetime.now().isoformat()
        
        self.state_manager.save(
            f"tool_exec:{execution_id}",
            execution.to_dict(),
            StateType.TOOL,
        )
        
        return True
    
    def get_execution(self, execution_id: str) -> Optional[ToolExecution]:
        """Get execution by ID."""
        data = self.state_manager.get(f"tool_exec:{execution_id}")
        if data:
            return ToolExecution.from_dict(data)
        return None
    
    def get_pending_executions(self, tool_name: str = None) -> List[ToolExecution]:
        """Get all pending/running executions."""
        keys = self.state_manager.list("tool_exec:")
        executions = []
        
        for key in keys:
            data = self.state_manager.get(key)
            if data:
                execution = ToolExecution.from_dict(data)
                if execution.status in [ToolExecutionStatus.PENDING, ToolExecutionStatus.RUNNING]:
                    if tool_name is None or execution.tool_name == tool_name:
                        executions.append(execution)
        
        return executions
    
    # Caching
    def cache_result(
        self,
        tool_name: str,
        args: Dict[str, Any],
        result: Any,
        ttl: int = None,
    ) -> str:
        """Cache tool result."""
        args_hash = self._hash_args(args)
        cache_key = f"tool_cache:{tool_name}:{args_hash}"
        
        expires_at = None
        if ttl:
            expires_at = (datetime.now() + timedelta(seconds=ttl)).isoformat()
        
        entry = ToolCacheEntry(
            cache_key=cache_key,
            tool_name=tool_name,
            args_hash=args_hash,
            result=result,
            expires_at=expires_at,
        )
        
        self.state_manager.save(cache_key, entry.to_dict(), StateType.TOOL)
        return cache_key
    
    def get_cached_result(
        self,
        tool_name: str,
        args: Dict[str, Any],
    ) -> Optional[Any]:
        """Get cached result if available."""
        args_hash = self._hash_args(args)
        cache_key = f"tool_cache:{tool_name}:{args_hash}"
        
        data = self.state_manager.get(cache_key)
        if not data:
            self._update_stats(tool_name, cache_miss=True)
            return None
        
        entry = ToolCacheEntry.from_dict(data)
        
        if entry.is_expired:
            self.state_manager.delete(cache_key)
            self._update_stats(tool_name, cache_miss=True)
            return None
        
        # Update hit count
        entry.hit_count += 1
        self.state_manager.save(cache_key, entry.to_dict(), StateType.TOOL)
        self._update_stats(tool_name, cache_hit=True)
        
        return entry.result
    
    def invalidate_cache(self, tool_name: str, args: Dict[str, Any] = None) -> int:
        """Invalidate cache entries."""
        if args:
            # Invalidate specific entry
            args_hash = self._hash_args(args)
            cache_key = f"tool_cache:{tool_name}:{args_hash}"
            return 1 if self.state_manager.delete(cache_key) else 0
        
        # Invalidate all entries for tool
        keys = self.state_manager.list(f"tool_cache:{tool_name}:")
        count = 0
        for key in keys:
            if self.state_manager.delete(key):
                count += 1
        return count
    
    def clear_expired_cache(self) -> int:
        """Clear all expired cache entries."""
        keys = self.state_manager.list("tool_cache:")
        count = 0
        
        for key in keys:
            data = self.state_manager.get(key)
            if data:
                entry = ToolCacheEntry.from_dict(data)
                if entry.is_expired:
                    if self.state_manager.delete(key):
                        count += 1
        
        return count
    
    # Retry Management
    def _schedule_retry(
        self,
        execution: ToolExecution,
        error: str,
    ) -> RetryState:
        """Schedule retry for failed execution."""
        attempt = execution.retry_count + 1
        
        # Exponential backoff
        backoff = self.default_backoff_base ** attempt
        next_retry = datetime.now() + timedelta(seconds=backoff)
        
        retry = RetryState(
            execution_id=execution.execution_id,
            tool_name=execution.tool_name,
            attempt=attempt,
            max_attempts=self.default_max_retries,
            next_retry_at=next_retry.isoformat(),
            backoff_seconds=backoff,
            last_error=error,
        )
        
        self.state_manager.save(
            f"tool_retry:{execution.execution_id}",
            retry.to_dict(),
            StateType.TOOL,
        )
        
        return retry
    
    def get_retry_state(self, execution_id: str) -> Optional[RetryState]:
        """Get retry state for execution."""
        data = self.state_manager.get(f"tool_retry:{execution_id}")
        if data:
            return RetryState.from_dict(data)
        return None
    
    def get_ready_retries(self) -> List[RetryState]:
        """Get all retry states that are ready to execute."""
        keys = self.state_manager.list("tool_retry:")
        ready = []
        
        for key in keys:
            data = self.state_manager.get(key)
            if data:
                retry = RetryState.from_dict(data)
                if retry.can_retry and retry.is_ready:
                    ready.append(retry)
        
        return ready
    
    def consume_retry(self, execution_id: str) -> Optional[ToolExecution]:
        """Consume retry and return execution for re-run."""
        retry = self.get_retry_state(execution_id)
        if not retry or not retry.can_retry or not retry.is_ready:
            return None
        
        # Get original execution
        data = self.state_manager.get(f"tool_exec:{execution_id}")
        if not data:
            return None
        
        execution = ToolExecution.from_dict(data)
        execution.status = ToolExecutionStatus.RETRYING
        execution.retry_count = retry.attempt
        execution.started_at = datetime.now().isoformat()
        execution.completed_at = None
        execution.result = None
        execution.error = None
        
        # Save updated execution
        self.state_manager.save(
            f"tool_exec:{execution_id}",
            execution.to_dict(),
            StateType.TOOL,
        )
        
        # Remove retry state
        self.state_manager.delete(f"tool_retry:{execution_id}")
        
        return execution
    
    def cancel_retry(self, execution_id: str) -> bool:
        """Cancel scheduled retry."""
        return self.state_manager.delete(f"tool_retry:{execution_id}")
    
    # Statistics
    def _update_stats(
        self,
        tool_name: str,
        success: bool = None,
        duration_ms: int = None,
        cache_hit: bool = False,
        cache_miss: bool = False,
    ):
        """Update tool statistics."""
        key = f"tool_stats:{tool_name}"
        data = self.state_manager.get(key)
        
        if data:
            stats = ToolStats.from_dict(data)
        else:
            stats = ToolStats(tool_name=tool_name)
        
        if success is not None:
            stats.total_executions += 1
            if success:
                stats.successful_executions += 1
            else:
                stats.failed_executions += 1
            
            if duration_ms:
                stats.total_duration_ms += duration_ms
                stats.avg_duration_ms = stats.total_duration_ms / stats.total_executions
            
            stats.last_executed = datetime.now().isoformat()
        
        if cache_hit:
            stats.cache_hits += 1
        if cache_miss:
            stats.cache_misses += 1
        
        self.state_manager.save(key, stats.to_dict(), StateType.TOOL)
    
    def get_stats(self, tool_name: str) -> Optional[ToolStats]:
        """Get statistics for a tool."""
        data = self.state_manager.get(f"tool_stats:{tool_name}")
        if data:
            return ToolStats.from_dict(data)
        return None
    
    def get_all_stats(self) -> List[ToolStats]:
        """Get statistics for all tools."""
        keys = self.state_manager.list("tool_stats:")
        stats = []
        
        for key in keys:
            data = self.state_manager.get(key)
            if data:
                stats.append(ToolStats.from_dict(data))
        
        return stats
    
    def reset_stats(self, tool_name: str = None) -> int:
        """Reset statistics."""
        if tool_name:
            return 1 if self.state_manager.delete(f"tool_stats:{tool_name}") else 0
        
        keys = self.state_manager.list("tool_stats:")
        count = 0
        for key in keys:
            if self.state_manager.delete(key):
                count += 1
        return count


__all__ = [
    "ToolExecutionStatus",
    "ToolExecution",
    "ToolCacheEntry",
    "RetryState",
    "ToolStats",
    "ToolStateManager",
]



"""
Enterprise Backup Manager Module.

Automated backups, restore operations,
snapshots, and disaster recovery.

Example:
    # Create backup manager
    manager = create_backup_manager()
    
    # Create backup
    backup = await manager.backup(
        source="/data/database",
        name="daily-backup",
        compress=True,
    )
    
    # List backups
    backups = await manager.list_backups()
    
    # Restore from backup
    await manager.restore(backup.id, target="/data/restore")
    
    # Schedule automated backups
    await manager.schedule("0 2 * * *", source="/data")
"""

from __future__ import annotations

import asyncio
import functools
import gzip
import hashlib
import io
import json
import logging
import os
import shutil
import tarfile
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class BackupError(Exception):
    """Backup error."""
    pass


class RestoreError(BackupError):
    """Restore error."""
    pass


class BackupNotFoundError(BackupError):
    """Backup not found."""
    pass


class BackupType(str, Enum):
    """Backup type."""
    FULL = "full"
    INCREMENTAL = "incremental"
    DIFFERENTIAL = "differential"
    SNAPSHOT = "snapshot"


class BackupStatus(str, Enum):
    """Backup status."""
    PENDING = "pending"
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"
    FAILED = "failed"
    EXPIRED = "expired"


class CompressionType(str, Enum):
    """Compression type."""
    NONE = "none"
    GZIP = "gzip"
    LZ4 = "lz4"
    ZSTD = "zstd"


@dataclass
class BackupMetadata:
    """Backup metadata."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    source: str = ""
    backup_type: BackupType = BackupType.FULL
    status: BackupStatus = BackupStatus.PENDING
    compression: CompressionType = CompressionType.GZIP
    size_bytes: int = 0
    file_count: int = 0
    checksum: Optional[str] = None
    created_at: datetime = field(default_factory=datetime.utcnow)
    completed_at: Optional[datetime] = None
    expires_at: Optional[datetime] = None
    retention_days: int = 30
    parent_backup_id: Optional[str] = None
    tags: List[str] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class RestoreResult:
    """Restore result."""
    success: bool
    backup_id: str
    target: str
    files_restored: int = 0
    size_bytes: int = 0
    started_at: datetime = field(default_factory=datetime.utcnow)
    completed_at: Optional[datetime] = None
    error: Optional[str] = None


@dataclass
class ScheduledBackup:
    """Scheduled backup configuration."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    cron_expression: str = ""
    source: str = ""
    name_template: str = "backup-{date}"
    backup_type: BackupType = BackupType.FULL
    compression: CompressionType = CompressionType.GZIP
    retention_days: int = 30
    enabled: bool = True
    last_run: Optional[datetime] = None
    next_run: Optional[datetime] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class BackupStats:
    """Backup statistics."""
    total_backups: int = 0
    total_size_bytes: int = 0
    completed_backups: int = 0
    failed_backups: int = 0
    oldest_backup: Optional[datetime] = None
    newest_backup: Optional[datetime] = None
    backups_by_type: Dict[str, int] = field(default_factory=dict)


# Storage backend
class BackupStorage(ABC):
    """Abstract backup storage."""
    
    @abstractmethod
    async def store(
        self,
        backup_id: str,
        data: BinaryIO,
    ) -> str:
        """Store backup data."""
        pass
    
    @abstractmethod
    async def retrieve(
        self,
        backup_id: str,
    ) -> BinaryIO:
        """Retrieve backup data."""
        pass
    
    @abstractmethod
    async def delete(self, backup_id: str) -> bool:
        """Delete backup."""
        pass
    
    @abstractmethod
    async def exists(self, backup_id: str) -> bool:
        """Check if backup exists."""
        pass


class LocalBackupStorage(BackupStorage):
    """Local filesystem backup storage."""
    
    def __init__(self, base_path: str = "/tmp/backups"):
        self._base_path = Path(base_path)
        self._base_path.mkdir(parents=True, exist_ok=True)
    
    async def store(
        self,
        backup_id: str,
        data: BinaryIO,
    ) -> str:
        """Store backup locally."""
        path = self._base_path / f"{backup_id}.backup"
        
        with open(path, "wb") as f:
            shutil.copyfileobj(data, f)
        
        return str(path)
    
    async def retrieve(
        self,
        backup_id: str,
    ) -> BinaryIO:
        """Retrieve backup."""
        path = self._base_path / f"{backup_id}.backup"
        
        if not path.exists():
            raise BackupNotFoundError(f"Backup not found: {backup_id}")
        
        return open(path, "rb")
    
    async def delete(self, backup_id: str) -> bool:
        """Delete backup."""
        path = self._base_path / f"{backup_id}.backup"
        
        if path.exists():
            path.unlink()
            return True
        return False
    
    async def exists(self, backup_id: str) -> bool:
        """Check if backup exists."""
        path = self._base_path / f"{backup_id}.backup"
        return path.exists()


class InMemoryBackupStorage(BackupStorage):
    """In-memory backup storage for testing."""
    
    def __init__(self):
        self._backups: Dict[str, bytes] = {}
    
    async def store(
        self,
        backup_id: str,
        data: BinaryIO,
    ) -> str:
        """Store backup."""
        self._backups[backup_id] = data.read()
        return backup_id
    
    async def retrieve(
        self,
        backup_id: str,
    ) -> BinaryIO:
        """Retrieve backup."""
        if backup_id not in self._backups:
            raise BackupNotFoundError(f"Backup not found: {backup_id}")
        
        return io.BytesIO(self._backups[backup_id])
    
    async def delete(self, backup_id: str) -> bool:
        """Delete backup."""
        if backup_id in self._backups:
            del self._backups[backup_id]
            return True
        return False
    
    async def exists(self, backup_id: str) -> bool:
        """Check if backup exists."""
        return backup_id in self._backups


# Metadata store
class MetadataStore(ABC):
    """Abstract metadata store."""
    
    @abstractmethod
    async def save(self, metadata: BackupMetadata) -> None:
        """Save metadata."""
        pass
    
    @abstractmethod
    async def get(self, backup_id: str) -> Optional[BackupMetadata]:
        """Get metadata."""
        pass
    
    @abstractmethod
    async def list(
        self,
        source: Optional[str] = None,
        status: Optional[BackupStatus] = None,
        limit: int = 100,
    ) -> List[BackupMetadata]:
        """List backups."""
        pass
    
    @abstractmethod
    async def delete(self, backup_id: str) -> bool:
        """Delete metadata."""
        pass


class InMemoryMetadataStore(MetadataStore):
    """In-memory metadata store."""
    
    def __init__(self):
        self._backups: Dict[str, BackupMetadata] = {}
    
    async def save(self, metadata: BackupMetadata) -> None:
        self._backups[metadata.id] = metadata
    
    async def get(self, backup_id: str) -> Optional[BackupMetadata]:
        return self._backups.get(backup_id)
    
    async def list(
        self,
        source: Optional[str] = None,
        status: Optional[BackupStatus] = None,
        limit: int = 100,
    ) -> List[BackupMetadata]:
        results = []
        
        for backup in self._backups.values():
            if source and backup.source != source:
                continue
            if status and backup.status != status:
                continue
            
            results.append(backup)
        
        results.sort(key=lambda b: b.created_at, reverse=True)
        return results[:limit]
    
    async def delete(self, backup_id: str) -> bool:
        if backup_id in self._backups:
            del self._backups[backup_id]
            return True
        return False


# Backup manager
class BackupManager:
    """
    Backup management service.
    """
    
    def __init__(
        self,
        storage: Optional[BackupStorage] = None,
        metadata_store: Optional[MetadataStore] = None,
        default_retention_days: int = 30,
    ):
        self._storage = storage or InMemoryBackupStorage()
        self._metadata = metadata_store or InMemoryMetadataStore()
        self._default_retention_days = default_retention_days
        self._scheduled: Dict[str, ScheduledBackup] = {}
        self._scheduler_task: Optional[asyncio.Task] = None
        self._running = False
        self._hooks: Dict[str, List[Callable]] = {}
    
    async def backup(
        self,
        source: str,
        name: Optional[str] = None,
        backup_type: BackupType = BackupType.FULL,
        compression: CompressionType = CompressionType.GZIP,
        retention_days: Optional[int] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> BackupMetadata:
        """
        Create backup.
        
        Args:
            source: Source path or data
            name: Backup name
            backup_type: Type of backup
            compression: Compression type
            retention_days: Retention period
            tags: Backup tags
            metadata: Additional metadata
            
        Returns:
            Backup metadata
        """
        backup_meta = BackupMetadata(
            name=name or f"backup-{datetime.utcnow().strftime('%Y%m%d-%H%M%S')}",
            source=source,
            backup_type=backup_type,
            compression=compression,
            retention_days=retention_days or self._default_retention_days,
            tags=tags or [],
            metadata=metadata or {},
        )
        
        backup_meta.status = BackupStatus.IN_PROGRESS
        await self._metadata.save(backup_meta)
        
        await self._trigger("backup_started", backup_meta)
        
        try:
            # Create backup data
            data, file_count = await self._create_backup_data(
                source, compression
            )
            
            # Calculate checksum
            data.seek(0)
            checksum = hashlib.sha256(data.read()).hexdigest()
            data.seek(0)
            
            # Store backup
            await self._storage.store(backup_meta.id, data)
            
            # Update metadata
            data.seek(0, 2)  # Seek to end
            backup_meta.size_bytes = data.tell()
            backup_meta.file_count = file_count
            backup_meta.checksum = checksum
            backup_meta.status = BackupStatus.COMPLETED
            backup_meta.completed_at = datetime.utcnow()
            backup_meta.expires_at = (
                datetime.utcnow() + timedelta(days=backup_meta.retention_days)
            )
            
            await self._metadata.save(backup_meta)
            await self._trigger("backup_completed", backup_meta)
            
            logger.info(f"Backup created: {backup_meta.name} ({backup_meta.id})")
            
            return backup_meta
            
        except Exception as e:
            backup_meta.status = BackupStatus.FAILED
            backup_meta.metadata["error"] = str(e)
            await self._metadata.save(backup_meta)
            
            await self._trigger("backup_failed", backup_meta, e)
            
            logger.error(f"Backup failed: {e}")
            raise BackupError(f"Backup failed: {e}")
    
    async def _create_backup_data(
        self,
        source: str,
        compression: CompressionType,
    ) -> Tuple[BinaryIO, int]:
        """Create backup data from source."""
        buffer = io.BytesIO()
        file_count = 0
        
        source_path = Path(source)
        
        if source_path.exists():
            # Create tar archive
            if compression == CompressionType.GZIP:
                mode = "w:gz"
            else:
                mode = "w"
            
            with tarfile.open(fileobj=buffer, mode=mode) as tar:
                if source_path.is_file():
                    tar.add(str(source_path), arcname=source_path.name)
                    file_count = 1
                else:
                    for item in source_path.rglob("*"):
                        if item.is_file():
                            tar.add(
                                str(item),
                                arcname=str(item.relative_to(source_path)),
                            )
                            file_count += 1
        else:
            # Treat as data string
            data = source.encode()
            if compression == CompressionType.GZIP:
                data = gzip.compress(data)
            buffer.write(data)
            file_count = 1
        
        buffer.seek(0)
        return buffer, file_count
    
    async def restore(
        self,
        backup_id: str,
        target: str,
        verify: bool = True,
    ) -> RestoreResult:
        """
        Restore from backup.
        
        Args:
            backup_id: Backup ID
            target: Target path
            verify: Verify checksum
            
        Returns:
            Restore result
        """
        result = RestoreResult(
            success=False,
            backup_id=backup_id,
            target=target,
        )
        
        try:
            metadata = await self._metadata.get(backup_id)
            
            if not metadata:
                raise BackupNotFoundError(f"Backup not found: {backup_id}")
            
            if metadata.status != BackupStatus.COMPLETED:
                raise RestoreError(f"Cannot restore incomplete backup")
            
            await self._trigger("restore_started", metadata, target)
            
            # Retrieve backup data
            data = await self._storage.retrieve(backup_id)
            
            # Verify checksum
            if verify and metadata.checksum:
                data_content = data.read()
                actual_checksum = hashlib.sha256(data_content).hexdigest()
                
                if actual_checksum != metadata.checksum:
                    raise RestoreError("Checksum verification failed")
                
                data = io.BytesIO(data_content)
            
            # Extract backup
            target_path = Path(target)
            target_path.mkdir(parents=True, exist_ok=True)
            
            if metadata.compression == CompressionType.GZIP:
                mode = "r:gz"
            else:
                mode = "r"
            
            try:
                with tarfile.open(fileobj=data, mode=mode) as tar:
                    tar.extractall(target_path)
                    result.files_restored = len(tar.getmembers())
            except tarfile.ReadError:
                # Not a tar file, extract raw data
                data.seek(0)
                content = data.read()
                
                if metadata.compression == CompressionType.GZIP:
                    content = gzip.decompress(content)
                
                output_file = target_path / "restored_data"
                output_file.write_bytes(content)
                result.files_restored = 1
            
            result.success = True
            result.completed_at = datetime.utcnow()
            result.size_bytes = metadata.size_bytes
            
            await self._trigger("restore_completed", metadata, result)
            
            logger.info(f"Restored backup {backup_id} to {target}")
            
            return result
            
        except Exception as e:
            result.error = str(e)
            result.completed_at = datetime.utcnow()
            
            logger.error(f"Restore failed: {e}")
            raise RestoreError(f"Restore failed: {e}")
    
    async def list_backups(
        self,
        source: Optional[str] = None,
        status: Optional[BackupStatus] = None,
        limit: int = 100,
    ) -> List[BackupMetadata]:
        """List backups."""
        return await self._metadata.list(source, status, limit)
    
    async def get_backup(self, backup_id: str) -> Optional[BackupMetadata]:
        """Get backup metadata."""
        return await self._metadata.get(backup_id)
    
    async def delete_backup(self, backup_id: str) -> bool:
        """Delete backup."""
        if await self._storage.delete(backup_id):
            await self._metadata.delete(backup_id)
            logger.info(f"Deleted backup: {backup_id}")
            return True
        return False
    
    async def expire_backups(self) -> int:
        """Expire old backups."""
        now = datetime.utcnow()
        expired_count = 0
        
        backups = await self._metadata.list(
            status=BackupStatus.COMPLETED,
            limit=10000,
        )
        
        for backup in backups:
            if backup.expires_at and backup.expires_at < now:
                backup.status = BackupStatus.EXPIRED
                await self._metadata.save(backup)
                await self._storage.delete(backup.id)
                expired_count += 1
        
        if expired_count:
            logger.info(f"Expired {expired_count} backups")
        
        return expired_count
    
    async def schedule(
        self,
        cron_expression: str,
        source: str,
        name_template: str = "backup-{date}",
        backup_type: BackupType = BackupType.FULL,
        retention_days: Optional[int] = None,
    ) -> ScheduledBackup:
        """
        Schedule automated backup.
        
        Args:
            cron_expression: Cron expression
            source: Source path
            name_template: Name template
            backup_type: Backup type
            retention_days: Retention days
            
        Returns:
            Scheduled backup config
        """
        scheduled = ScheduledBackup(
            cron_expression=cron_expression,
            source=source,
            name_template=name_template,
            backup_type=backup_type,
            retention_days=retention_days or self._default_retention_days,
        )
        
        self._scheduled[scheduled.id] = scheduled
        
        logger.info(f"Scheduled backup: {cron_expression}")
        
        return scheduled
    
    async def unschedule(self, schedule_id: str) -> bool:
        """Remove scheduled backup."""
        if schedule_id in self._scheduled:
            del self._scheduled[schedule_id]
            return True
        return False
    
    async def start_scheduler(self) -> None:
        """Start backup scheduler."""
        self._running = True
        self._scheduler_task = asyncio.create_task(self._scheduler_loop())
        logger.info("Backup scheduler started")
    
    async def stop_scheduler(self) -> None:
        """Stop backup scheduler."""
        self._running = False
        if self._scheduler_task:
            self._scheduler_task.cancel()
        logger.info("Backup scheduler stopped")
    
    async def _scheduler_loop(self) -> None:
        """Scheduler loop."""
        while self._running:
            try:
                await self._check_scheduled_backups()
                await self.expire_backups()
                await asyncio.sleep(60)  # Check every minute
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error(f"Scheduler error: {e}")
    
    async def _check_scheduled_backups(self) -> None:
        """Check and run scheduled backups."""
        now = datetime.utcnow()
        
        for scheduled in self._scheduled.values():
            if not scheduled.enabled:
                continue
            
            # Simple minute-based check
            if scheduled.next_run and now >= scheduled.next_run:
                try:
                    name = scheduled.name_template.format(
                        date=now.strftime("%Y%m%d-%H%M%S")
                    )
                    
                    await self.backup(
                        source=scheduled.source,
                        name=name,
                        backup_type=scheduled.backup_type,
                        retention_days=scheduled.retention_days,
                    )
                    
                    scheduled.last_run = now
                    # Set next run (simplified - add 1 day)
                    scheduled.next_run = now + timedelta(days=1)
                    
                except Exception as e:
                    logger.error(f"Scheduled backup failed: {e}")
    
    def on(self, event: str, handler: Callable) -> None:
        """Add event handler."""
        if event not in self._hooks:
            self._hooks[event] = []
        self._hooks[event].append(handler)
    
    async def _trigger(self, event: str, *args, **kwargs) -> None:
        """Trigger event."""
        for handler in self._hooks.get(event, []):
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(*args, **kwargs)
                else:
                    handler(*args, **kwargs)
            except Exception as e:
                logger.error(f"Hook error: {e}")
    
    async def get_stats(self) -> BackupStats:
        """Get backup statistics."""
        backups = await self._metadata.list(limit=10000)
        
        stats = BackupStats()
        
        for backup in backups:
            stats.total_backups += 1
            stats.total_size_bytes += backup.size_bytes
            
            if backup.status == BackupStatus.COMPLETED:
                stats.completed_backups += 1
            elif backup.status == BackupStatus.FAILED:
                stats.failed_backups += 1
            
            type_key = backup.backup_type.value
            stats.backups_by_type[type_key] = (
                stats.backups_by_type.get(type_key, 0) + 1
            )
            
            if not stats.oldest_backup or backup.created_at < stats.oldest_backup:
                stats.oldest_backup = backup.created_at
            if not stats.newest_backup or backup.created_at > stats.newest_backup:
                stats.newest_backup = backup.created_at
        
        return stats


# Factory functions
def create_backup_manager(
    storage_path: Optional[str] = None,
    retention_days: int = 30,
) -> BackupManager:
    """Create backup manager."""
    if storage_path:
        storage = LocalBackupStorage(storage_path)
    else:
        storage = InMemoryBackupStorage()
    
    return BackupManager(
        storage=storage,
        default_retention_days=retention_days,
    )


def create_local_storage(path: str) -> LocalBackupStorage:
    """Create local storage."""
    return LocalBackupStorage(path)


__all__ = [
    # Exceptions
    "BackupError",
    "RestoreError",
    "BackupNotFoundError",
    # Enums
    "BackupType",
    "BackupStatus",
    "CompressionType",
    # Data classes
    "BackupMetadata",
    "RestoreResult",
    "ScheduledBackup",
    "BackupStats",
    # Storage
    "BackupStorage",
    "LocalBackupStorage",
    "InMemoryBackupStorage",
    # Metadata
    "MetadataStore",
    "InMemoryMetadataStore",
    # Manager
    "BackupManager",
    # Factory functions
    "create_backup_manager",
    "create_local_storage",
]



"""
Enterprise JSON Mode - Structured JSON output with validation.

Provides guaranteed JSON output from LLMs with
Pydantic model validation and schema generation.

Features:
- Pydantic model integration
- JSON schema generation
- Output validation
- Auto-repair for malformed JSON
- Type-safe responses
"""

import json
import logging
import re
from dataclasses import dataclass, field
from typing import (
    Any, Callable, Dict, Generic, List, Optional, 
    Type, TypeVar, Union, get_args, get_origin,
)

logger = logging.getLogger(__name__)


# Type variable for generic model
T = TypeVar("T")


# =============================================================================
# JSON Schema Generation
# =============================================================================

def python_type_to_json_schema(python_type: Type) -> Dict[str, Any]:
    """Convert Python type to JSON schema."""
    origin = get_origin(python_type)
    args = get_args(python_type)
    
    # Handle None
    if python_type is type(None):
        return {"type": "null"}
    
    # Handle basic types
    if python_type is str:
        return {"type": "string"}
    elif python_type is int:
        return {"type": "integer"}
    elif python_type is float:
        return {"type": "number"}
    elif python_type is bool:
        return {"type": "boolean"}
    
    # Handle Optional
    if origin is Union:
        non_none_types = [t for t in args if t is not type(None)]
        if len(non_none_types) == 1:
            # Optional[X] -> nullable X
            schema = python_type_to_json_schema(non_none_types[0])
            return {"anyOf": [schema, {"type": "null"}]}
        else:
            # Union of multiple types
            return {"anyOf": [python_type_to_json_schema(t) for t in non_none_types]}
    
    # Handle List
    if origin is list:
        if args:
            return {
                "type": "array",
                "items": python_type_to_json_schema(args[0]),
            }
        return {"type": "array"}
    
    # Handle Dict
    if origin is dict:
        return {"type": "object"}
    
    # Handle Pydantic models
    try:
        from pydantic import BaseModel
        if isinstance(python_type, type) and issubclass(python_type, BaseModel):
            return python_type.model_json_schema()
    except ImportError:
        pass
    
    # Handle dataclasses
    import dataclasses
    if dataclasses.is_dataclass(python_type):
        properties = {}
        required = []
        
        for f in dataclasses.fields(python_type):
            properties[f.name] = python_type_to_json_schema(f.type)
            if f.default is dataclasses.MISSING and f.default_factory is dataclasses.MISSING:
                required.append(f.name)
        
        return {
            "type": "object",
            "properties": properties,
            "required": required,
        }
    
    # Fallback
    return {"type": "object"}


def generate_json_schema(model: Type[T]) -> Dict[str, Any]:
    """Generate JSON schema from a model."""
    try:
        from pydantic import BaseModel
        if isinstance(model, type) and issubclass(model, BaseModel):
            return model.model_json_schema()
    except ImportError:
        pass
    
    return python_type_to_json_schema(model)


# =============================================================================
# JSON Extractor
# =============================================================================

class JSONExtractor:
    """
    Extracts JSON from LLM responses.
    
    Handles common issues like:
    - Markdown code blocks
    - Leading/trailing text
    - Malformed JSON
    """
    
    # Patterns for extracting JSON
    CODE_BLOCK_PATTERN = re.compile(r'```(?:json)?\s*([\s\S]*?)```')
    JSON_OBJECT_PATTERN = re.compile(r'\{[\s\S]*\}')
    JSON_ARRAY_PATTERN = re.compile(r'\[[\s\S]*\]')
    
    def extract(self, text: str) -> str:
        """Extract JSON from text."""
        # Try code block first
        match = self.CODE_BLOCK_PATTERN.search(text)
        if match:
            return match.group(1).strip()
        
        # Try to find JSON object
        match = self.JSON_OBJECT_PATTERN.search(text)
        if match:
            return match.group(0)
        
        # Try to find JSON array
        match = self.JSON_ARRAY_PATTERN.search(text)
        if match:
            return match.group(0)
        
        # Return original text
        return text.strip()
    
    def extract_and_parse(self, text: str) -> Any:
        """Extract and parse JSON from text."""
        json_str = self.extract(text)
        return json.loads(json_str)


class JSONRepairer:
    """
    Attempts to repair malformed JSON.
    
    Handles common issues:
    - Missing quotes
    - Trailing commas
    - Single quotes instead of double
    - Missing closing brackets
    """
    
    def repair(self, text: str) -> str:
        """Attempt to repair malformed JSON."""
        # Replace single quotes with double quotes
        text = re.sub(r"(?<!\\)'", '"', text)
        
        # Remove trailing commas
        text = re.sub(r',\s*([}\]])', r'\1', text)
        
        # Add missing closing brackets
        open_braces = text.count('{') - text.count('}')
        open_brackets = text.count('[') - text.count(']')
        
        if open_braces > 0:
            text += '}' * open_braces
        
        if open_brackets > 0:
            text += ']' * open_brackets
        
        # Fix unquoted keys
        text = re.sub(r'(\{|,)\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*:', r'\1"\2":', text)
        
        return text
    
    def try_repair(self, text: str) -> Optional[Any]:
        """Try to repair and parse JSON."""
        repaired = self.repair(text)
        
        try:
            return json.loads(repaired)
        except json.JSONDecodeError:
            return None


# =============================================================================
# Typed Response
# =============================================================================

@dataclass
class JSONResponse(Generic[T]):
    """Response with parsed JSON."""
    raw_response: str
    json_str: str
    parsed: T
    
    # Validation
    is_valid: bool = True
    validation_errors: List[str] = field(default_factory=list)
    
    # Repair info
    was_repaired: bool = False


# =============================================================================
# JSON Mode Handler
# =============================================================================

class JSONMode(Generic[T]):
    """
    Handles JSON mode for LLM responses.
    
    Usage:
        >>> from pydantic import BaseModel
        >>> 
        >>> class User(BaseModel):
        ...     name: str
        ...     age: int
        >>> 
        >>> json_mode = JSONMode(User)
        >>> 
        >>> # Get system prompt for JSON mode
        >>> system_prompt = json_mode.get_system_prompt()
        >>> 
        >>> # Parse response
        >>> response = json_mode.parse(llm_response)
        >>> print(response.parsed.name)
    """
    
    def __init__(
        self,
        model: Type[T],
        strict: bool = True,
        auto_repair: bool = True,
    ):
        self.model = model
        self.strict = strict
        self.auto_repair = auto_repair
        
        self._extractor = JSONExtractor()
        self._repairer = JSONRepairer()
        self._schema = generate_json_schema(model)
    
    @property
    def schema(self) -> Dict[str, Any]:
        """Get the JSON schema."""
        return self._schema
    
    def get_system_prompt(self, base_prompt: str = "") -> str:
        """Get system prompt with JSON schema."""
        schema_str = json.dumps(self._schema, indent=2)
        
        json_instruction = f"""
{base_prompt}

You must respond with valid JSON that matches this schema:
```json
{schema_str}
```

Important:
- Respond ONLY with the JSON object, no additional text
- Ensure all required fields are present
- Use the correct data types as specified
"""
        return json_instruction.strip()
    
    def get_function_schema(self) -> Dict[str, Any]:
        """Get schema for function calling."""
        return {
            "name": "output",
            "description": "The structured output",
            "parameters": self._schema,
        }
    
    def parse(self, response: str) -> JSONResponse[T]:
        """Parse LLM response to typed model."""
        try:
            # Extract JSON
            json_str = self._extractor.extract(response)
            
            # Try to parse
            try:
                data = json.loads(json_str)
                was_repaired = False
            except json.JSONDecodeError:
                if self.auto_repair:
                    data = self._repairer.try_repair(json_str)
                    if data is None:
                        raise
                    was_repaired = True
                else:
                    raise
            
            # Validate and create model
            parsed, errors = self._validate_and_create(data)
            
            return JSONResponse(
                raw_response=response,
                json_str=json_str,
                parsed=parsed,
                is_valid=len(errors) == 0,
                validation_errors=errors,
                was_repaired=was_repaired,
            )
            
        except Exception as e:
            logger.error(f"Failed to parse JSON response: {e}")
            
            return JSONResponse(
                raw_response=response,
                json_str="",
                parsed=None,
                is_valid=False,
                validation_errors=[str(e)],
            )
    
    def _validate_and_create(self, data: Any) -> tuple[Optional[T], List[str]]:
        """Validate data and create model instance."""
        errors = []
        
        try:
            # Try Pydantic
            from pydantic import BaseModel, ValidationError
            
            if isinstance(self.model, type) and issubclass(self.model, BaseModel):
                try:
                    return self.model.model_validate(data), []
                except ValidationError as e:
                    errors = [str(err) for err in e.errors()]
                    if self.strict:
                        return None, errors
                    # Try to create anyway
                    return self.model.model_construct(**data), errors
        except ImportError:
            pass
        
        # Try dataclass
        import dataclasses
        if dataclasses.is_dataclass(self.model):
            try:
                return self.model(**data), []
            except TypeError as e:
                return None, [str(e)]
        
        # Just return the data
        return data, []
    
    async def parse_async(self, response: str) -> JSONResponse[T]:
        """Async parse (same as sync for now)."""
        return self.parse(response)


# =============================================================================
# Decorator for JSON Mode
# =============================================================================

def json_output(model: Type[T], strict: bool = True):
    """
    Decorator to enable JSON mode for a function.
    
    Usage:
        >>> @json_output(User)
        ... async def get_user(prompt: str) -> User:
        ...     response = await llm.chat(prompt)
        ...     return response  # Will be parsed as JSON
    """
    def decorator(func: Callable) -> Callable:
        json_mode = JSONMode(model, strict=strict)
        
        async def async_wrapper(*args, **kwargs):
            result = await func(*args, **kwargs)
            if isinstance(result, str):
                response = json_mode.parse(result)
                if response.is_valid:
                    return response.parsed
                raise ValueError(f"Invalid JSON: {response.validation_errors}")
            return result
        
        def sync_wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, str):
                response = json_mode.parse(result)
                if response.is_valid:
                    return response.parsed
                raise ValueError(f"Invalid JSON: {response.validation_errors}")
            return result
        
        import asyncio
        if asyncio.iscoroutinefunction(func):
            async_wrapper.__name__ = func.__name__
            async_wrapper.__doc__ = func.__doc__
            async_wrapper.json_mode = json_mode
            return async_wrapper
        else:
            sync_wrapper.__name__ = func.__name__
            sync_wrapper.__doc__ = func.__doc__
            sync_wrapper.json_mode = json_mode
            return sync_wrapper
    
    return decorator


# =============================================================================
# Common Models
# =============================================================================

@dataclass
class TextWithConfidence:
    """Text with confidence score."""
    text: str
    confidence: float


@dataclass
class EntityExtraction:
    """Extracted entities."""
    entities: List[Dict[str, Any]]


@dataclass
class Classification:
    """Classification result."""
    label: str
    confidence: float
    reasoning: Optional[str] = None


@dataclass
class Sentiment:
    """Sentiment analysis result."""
    sentiment: str  # positive, negative, neutral
    score: float  # -1 to 1
    aspects: List[Dict[str, Any]] = field(default_factory=list)


@dataclass
class Summary:
    """Summary result."""
    summary: str
    key_points: List[str] = field(default_factory=list)
    word_count: int = 0


@dataclass
class Translation:
    """Translation result."""
    translated_text: str
    source_language: str
    target_language: str


@dataclass
class CodeGeneration:
    """Code generation result."""
    code: str
    language: str
    explanation: Optional[str] = None
    imports: List[str] = field(default_factory=list)


@dataclass
class StructuredAnswer:
    """Structured Q&A answer."""
    answer: str
    confidence: float
    sources: List[str] = field(default_factory=list)
    follow_up_questions: List[str] = field(default_factory=list)


# =============================================================================
# Helper Functions
# =============================================================================

def create_json_mode(model: Type[T], **kwargs) -> JSONMode[T]:
    """Create a JSON mode handler."""
    return JSONMode(model, **kwargs)


def extract_json(text: str) -> Any:
    """Extract and parse JSON from text."""
    return JSONExtractor().extract_and_parse(text)


def repair_json(text: str) -> str:
    """Attempt to repair malformed JSON."""
    return JSONRepairer().repair(text)


def schema_from_model(model: Type) -> Dict[str, Any]:
    """Generate JSON schema from a model."""
    return generate_json_schema(model)



"""
Enterprise License Manager Module.

License tracking, usage monitoring, compliance,
and license expiration management.

Example:
    # Create license manager
    licenses = create_license_manager()
    
    # Add license
    license = await licenses.add(
        name="enterprise_api",
        license_type=LicenseType.SUBSCRIPTION,
        seats=100,
        expires_at=datetime(2025, 12, 31),
    )
    
    # Check out seat
    await licenses.checkout(
        license_name="enterprise_api",
        user_id="user123",
    )
    
    # Check compliance
    report = await licenses.check_compliance()
"""

from __future__ import annotations

import asyncio
import hashlib
import logging
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

logger = logging.getLogger(__name__)


class LicenseError(Exception):
    """License error."""
    pass


class LicenseExpired(LicenseError):
    """License expired."""
    pass


class NoSeatsAvailable(LicenseError):
    """No seats available."""
    pass


class LicenseType(str, Enum):
    """License type."""
    PERPETUAL = "perpetual"
    SUBSCRIPTION = "subscription"
    TRIAL = "trial"
    EVALUATION = "evaluation"
    OPEN_SOURCE = "open_source"
    CONCURRENT = "concurrent"
    NAMED_USER = "named_user"
    SITE = "site"


class LicenseStatus(str, Enum):
    """License status."""
    ACTIVE = "active"
    EXPIRED = "expired"
    SUSPENDED = "suspended"
    REVOKED = "revoked"
    PENDING = "pending"


class UsageEventType(str, Enum):
    """Usage event type."""
    CHECKOUT = "checkout"
    CHECKIN = "checkin"
    ACCESS = "access"
    FEATURE_USE = "feature_use"
    OVERAGE = "overage"


@dataclass
class UsageEvent:
    """License usage event."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    license_id: str = ""
    event_type: UsageEventType = UsageEventType.ACCESS
    user_id: str = ""
    feature: str = ""
    quantity: int = 1
    timestamp: datetime = field(default_factory=datetime.utcnow)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class LicenseKey:
    """License key."""
    key: str = ""
    created_at: datetime = field(default_factory=datetime.utcnow)
    activated_at: Optional[datetime] = None
    machine_id: str = ""
    valid: bool = True


@dataclass
class License:
    """License."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    license_type: LicenseType = LicenseType.SUBSCRIPTION
    status: LicenseStatus = LicenseStatus.ACTIVE
    
    # Keys
    license_key: Optional[LicenseKey] = None
    
    # Limits
    seats: Optional[int] = None  # None = unlimited
    max_usage: Optional[int] = None
    
    # Features
    features: List[str] = field(default_factory=list)
    restrictions: Dict[str, Any] = field(default_factory=dict)
    
    # Users
    active_users: Set[str] = field(default_factory=set)
    
    # Dates
    issued_at: datetime = field(default_factory=datetime.utcnow)
    starts_at: datetime = field(default_factory=datetime.utcnow)
    expires_at: Optional[datetime] = None
    
    # Vendor
    vendor: str = ""
    product: str = ""
    version: str = ""
    
    # Usage
    usage_count: int = 0
    
    # Metadata
    tags: Dict[str, str] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    @property
    def is_expired(self) -> bool:
        if self.expires_at is None:
            return False
        return datetime.utcnow() > self.expires_at
    
    @property
    def is_active(self) -> bool:
        return self.status == LicenseStatus.ACTIVE and not self.is_expired
    
    @property
    def available_seats(self) -> Optional[int]:
        if self.seats is None:
            return None
        return max(0, self.seats - len(self.active_users))
    
    @property
    def days_until_expiry(self) -> Optional[int]:
        if self.expires_at is None:
            return None
        delta = self.expires_at - datetime.utcnow()
        return max(0, delta.days)


@dataclass
class ComplianceIssue:
    """Compliance issue."""
    license_id: str = ""
    license_name: str = ""
    issue_type: str = ""
    severity: str = "warning"
    message: str = ""
    detected_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class ComplianceReport:
    """Compliance report."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    generated_at: datetime = field(default_factory=datetime.utcnow)
    total_licenses: int = 0
    active_licenses: int = 0
    expired_licenses: int = 0
    compliant: bool = True
    issues: List[ComplianceIssue] = field(default_factory=list)
    usage_summary: Dict[str, Any] = field(default_factory=dict)


@dataclass
class LicenseStats:
    """License statistics."""
    total_licenses: int = 0
    active_licenses: int = 0
    expired_licenses: int = 0
    total_seats: int = 0
    used_seats: int = 0
    by_type: Dict[str, int] = field(default_factory=dict)
    expiring_soon: List[str] = field(default_factory=list)


# License store
class LicenseStore(ABC):
    """License storage."""
    
    @abstractmethod
    async def save(self, license: License) -> None:
        pass
    
    @abstractmethod
    async def get(self, license_id: str) -> Optional[License]:
        pass
    
    @abstractmethod
    async def get_by_name(self, name: str) -> Optional[License]:
        pass
    
    @abstractmethod
    async def list_all(self) -> List[License]:
        pass
    
    @abstractmethod
    async def delete(self, license_id: str) -> bool:
        pass


class InMemoryLicenseStore(LicenseStore):
    """In-memory license store."""
    
    def __init__(self):
        self._licenses: Dict[str, License] = {}
        self._by_name: Dict[str, str] = {}
    
    async def save(self, license: License) -> None:
        self._licenses[license.id] = license
        self._by_name[license.name] = license.id
    
    async def get(self, license_id: str) -> Optional[License]:
        return self._licenses.get(license_id)
    
    async def get_by_name(self, name: str) -> Optional[License]:
        license_id = self._by_name.get(name)
        if license_id:
            return self._licenses.get(license_id)
        return None
    
    async def list_all(self) -> List[License]:
        return list(self._licenses.values())
    
    async def delete(self, license_id: str) -> bool:
        license = self._licenses.get(license_id)
        if license:
            del self._licenses[license_id]
            self._by_name.pop(license.name, None)
            return True
        return False


# Usage store
class UsageStore(ABC):
    """Usage event storage."""
    
    @abstractmethod
    async def save(self, event: UsageEvent) -> None:
        pass
    
    @abstractmethod
    async def list_by_license(
        self,
        license_id: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> List[UsageEvent]:
        pass
    
    @abstractmethod
    async def get_usage_count(
        self,
        license_id: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> int:
        pass


class InMemoryUsageStore(UsageStore):
    """In-memory usage store."""
    
    def __init__(self, max_events: int = 100000):
        self._events: Dict[str, UsageEvent] = {}
        self._by_license: Dict[str, List[str]] = {}
        self._max_events = max_events
    
    async def save(self, event: UsageEvent) -> None:
        self._events[event.id] = event
        
        if event.license_id not in self._by_license:
            self._by_license[event.license_id] = []
        
        self._by_license[event.license_id].append(event.id)
        
        # Trim
        if len(self._events) > self._max_events:
            oldest = list(self._events.keys())[0]
            del self._events[oldest]
    
    async def list_by_license(
        self,
        license_id: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> List[UsageEvent]:
        event_ids = self._by_license.get(license_id, [])
        events = []
        
        for event_id in event_ids:
            event = self._events.get(event_id)
            if event:
                if start_date and event.timestamp < start_date:
                    continue
                if end_date and event.timestamp > end_date:
                    continue
                events.append(event)
        
        return sorted(events, key=lambda e: e.timestamp, reverse=True)
    
    async def get_usage_count(
        self,
        license_id: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> int:
        events = await self.list_by_license(license_id, start_date, end_date)
        return sum(e.quantity for e in events)


# License key generator
class LicenseKeyGenerator:
    """License key generator."""
    
    @staticmethod
    def generate(
        license_id: str,
        secret: str = "default_secret",
        segments: int = 4,
        segment_length: int = 4,
    ) -> str:
        """Generate license key."""
        data = f"{license_id}:{secret}:{time.time()}"
        hash_bytes = hashlib.sha256(data.encode()).hexdigest()
        
        key_parts = []
        for i in range(segments):
            start = i * segment_length
            key_parts.append(hash_bytes[start:start + segment_length].upper())
        
        return "-".join(key_parts)
    
    @staticmethod
    def validate(
        key: str,
        min_segments: int = 4,
        segment_length: int = 4,
    ) -> bool:
        """Validate license key format."""
        parts = key.split("-")
        
        if len(parts) < min_segments:
            return False
        
        for part in parts:
            if len(part) != segment_length:
                return False
            if not all(c.isalnum() for c in part):
                return False
        
        return True


# License manager
class LicenseManager:
    """License manager."""
    
    def __init__(
        self,
        license_store: Optional[LicenseStore] = None,
        usage_store: Optional[UsageStore] = None,
    ):
        self._license_store = license_store or InMemoryLicenseStore()
        self._usage_store = usage_store or InMemoryUsageStore()
        self._key_generator = LicenseKeyGenerator()
        self._alert_handlers: List[Callable] = []
        self._expiry_threshold_days = 30
    
    async def add(
        self,
        name: str,
        license_type: Union[str, LicenseType] = LicenseType.SUBSCRIPTION,
        seats: Optional[int] = None,
        expires_at: Optional[datetime] = None,
        features: Optional[List[str]] = None,
        generate_key: bool = True,
        **metadata,
    ) -> License:
        """Add license."""
        if isinstance(license_type, str):
            license_type = LicenseType(license_type)
        
        license = License(
            name=name,
            license_type=license_type,
            seats=seats,
            expires_at=expires_at,
            features=features or [],
            metadata=metadata,
        )
        
        if generate_key:
            key = self._key_generator.generate(license.id)
            license.license_key = LicenseKey(
                key=key,
                activated_at=datetime.utcnow(),
            )
        
        await self._license_store.save(license)
        
        logger.info(f"License added: {name} ({license_type.value})")
        
        return license
    
    async def get(self, name: str) -> Optional[License]:
        """Get license by name."""
        return await self._license_store.get_by_name(name)
    
    async def get_by_id(self, license_id: str) -> Optional[License]:
        """Get license by ID."""
        return await self._license_store.get(license_id)
    
    async def update(
        self,
        name: str,
        **updates,
    ) -> Optional[License]:
        """Update license."""
        license = await self._license_store.get_by_name(name)
        
        if not license:
            return None
        
        for key, value in updates.items():
            if hasattr(license, key):
                setattr(license, key, value)
        
        await self._license_store.save(license)
        
        return license
    
    async def delete(self, name: str) -> bool:
        """Delete license."""
        license = await self._license_store.get_by_name(name)
        
        if license:
            return await self._license_store.delete(license.id)
        
        return False
    
    async def list_licenses(self) -> List[License]:
        """List all licenses."""
        return await self._license_store.list_all()
    
    async def checkout(
        self,
        license_name: str,
        user_id: str,
        feature: str = "",
    ) -> bool:
        """Check out license seat."""
        license = await self._license_store.get_by_name(license_name)
        
        if not license:
            raise LicenseError(f"License not found: {license_name}")
        
        if not license.is_active:
            if license.is_expired:
                raise LicenseExpired(f"License expired: {license_name}")
            raise LicenseError(f"License not active: {license_name}")
        
        # Check seats
        if license.seats is not None:
            if user_id not in license.active_users:
                if license.available_seats is not None and license.available_seats <= 0:
                    raise NoSeatsAvailable(f"No seats available: {license_name}")
        
        # Add user
        license.active_users.add(user_id)
        license.usage_count += 1
        
        await self._license_store.save(license)
        
        # Record usage
        event = UsageEvent(
            license_id=license.id,
            event_type=UsageEventType.CHECKOUT,
            user_id=user_id,
            feature=feature,
        )
        await self._usage_store.save(event)
        
        logger.debug(f"License checkout: {license_name} -> {user_id}")
        
        return True
    
    async def checkin(
        self,
        license_name: str,
        user_id: str,
    ) -> bool:
        """Check in license seat."""
        license = await self._license_store.get_by_name(license_name)
        
        if not license:
            return False
        
        if user_id in license.active_users:
            license.active_users.remove(user_id)
            await self._license_store.save(license)
            
            event = UsageEvent(
                license_id=license.id,
                event_type=UsageEventType.CHECKIN,
                user_id=user_id,
            )
            await self._usage_store.save(event)
            
            logger.debug(f"License checkin: {license_name} <- {user_id}")
            
            return True
        
        return False
    
    async def verify(
        self,
        license_name: str,
        user_id: Optional[str] = None,
        feature: Optional[str] = None,
    ) -> bool:
        """Verify license is valid."""
        license = await self._license_store.get_by_name(license_name)
        
        if not license:
            return False
        
        if not license.is_active:
            return False
        
        # Check feature
        if feature and license.features:
            if feature not in license.features:
                return False
        
        # Check user for named user license
        if license.license_type == LicenseType.NAMED_USER:
            if user_id and user_id not in license.active_users:
                return False
        
        return True
    
    async def has_feature(
        self,
        license_name: str,
        feature: str,
    ) -> bool:
        """Check if license has feature."""
        license = await self._license_store.get_by_name(license_name)
        
        if not license or not license.is_active:
            return False
        
        if not license.features:
            return True  # No restrictions
        
        return feature in license.features
    
    async def record_usage(
        self,
        license_name: str,
        user_id: str = "",
        feature: str = "",
        quantity: int = 1,
    ) -> None:
        """Record usage event."""
        license = await self._license_store.get_by_name(license_name)
        
        if not license:
            return
        
        event = UsageEvent(
            license_id=license.id,
            event_type=UsageEventType.ACCESS,
            user_id=user_id,
            feature=feature,
            quantity=quantity,
        )
        
        await self._usage_store.save(event)
        
        # Update usage count
        license.usage_count += quantity
        await self._license_store.save(license)
    
    async def get_usage(
        self,
        license_name: str,
        period_days: int = 30,
    ) -> Dict[str, Any]:
        """Get license usage."""
        license = await self._license_store.get_by_name(license_name)
        
        if not license:
            return {}
        
        start_date = datetime.utcnow() - timedelta(days=period_days)
        events = await self._usage_store.list_by_license(
            license.id,
            start_date=start_date,
        )
        
        by_type: Dict[str, int] = {}
        by_user: Dict[str, int] = {}
        by_feature: Dict[str, int] = {}
        
        for event in events:
            # By type
            event_type = event.event_type.value
            by_type[event_type] = by_type.get(event_type, 0) + event.quantity
            
            # By user
            if event.user_id:
                by_user[event.user_id] = by_user.get(event.user_id, 0) + event.quantity
            
            # By feature
            if event.feature:
                by_feature[event.feature] = by_feature.get(event.feature, 0) + event.quantity
        
        return {
            "license_name": license_name,
            "period_days": period_days,
            "total_events": len(events),
            "total_usage": sum(e.quantity for e in events),
            "by_type": by_type,
            "by_user": by_user,
            "by_feature": by_feature,
            "active_users": list(license.active_users),
        }
    
    async def check_compliance(self) -> ComplianceReport:
        """Check license compliance."""
        licenses = await self._license_store.list_all()
        
        report = ComplianceReport(
            total_licenses=len(licenses),
        )
        
        now = datetime.utcnow()
        
        for license in licenses:
            # Check active
            if license.is_active:
                report.active_licenses += 1
            
            # Check expired
            if license.is_expired:
                report.expired_licenses += 1
                report.issues.append(ComplianceIssue(
                    license_id=license.id,
                    license_name=license.name,
                    issue_type="expired",
                    severity="critical",
                    message=f"License expired on {license.expires_at}",
                ))
            
            # Check expiring soon
            elif license.days_until_expiry is not None:
                if license.days_until_expiry <= self._expiry_threshold_days:
                    report.issues.append(ComplianceIssue(
                        license_id=license.id,
                        license_name=license.name,
                        issue_type="expiring_soon",
                        severity="warning",
                        message=f"License expires in {license.days_until_expiry} days",
                    ))
            
            # Check seat overuse
            if license.seats is not None:
                used = len(license.active_users)
                if used > license.seats:
                    report.issues.append(ComplianceIssue(
                        license_id=license.id,
                        license_name=license.name,
                        issue_type="seat_overage",
                        severity="critical",
                        message=f"Using {used} seats but only {license.seats} licensed",
                    ))
            
            # Check usage limits
            if license.max_usage is not None:
                if license.usage_count > license.max_usage:
                    report.issues.append(ComplianceIssue(
                        license_id=license.id,
                        license_name=license.name,
                        issue_type="usage_overage",
                        severity="warning",
                        message=f"Usage {license.usage_count} exceeds limit {license.max_usage}",
                    ))
        
        # Determine overall compliance
        critical_issues = [i for i in report.issues if i.severity == "critical"]
        report.compliant = len(critical_issues) == 0
        
        return report
    
    async def get_expiring_licenses(
        self,
        days: int = 30,
    ) -> List[License]:
        """Get licenses expiring within days."""
        licenses = await self._license_store.list_all()
        expiring = []
        
        for license in licenses:
            if license.days_until_expiry is not None:
                if license.days_until_expiry <= days:
                    expiring.append(license)
        
        return sorted(expiring, key=lambda l: l.expires_at or datetime.max)
    
    async def get_stats(self) -> LicenseStats:
        """Get license statistics."""
        licenses = await self._license_store.list_all()
        
        stats = LicenseStats(
            total_licenses=len(licenses),
        )
        
        for license in licenses:
            # By status
            if license.is_active:
                stats.active_licenses += 1
            if license.is_expired:
                stats.expired_licenses += 1
            
            # By type
            ltype = license.license_type.value
            stats.by_type[ltype] = stats.by_type.get(ltype, 0) + 1
            
            # Seats
            if license.seats:
                stats.total_seats += license.seats
                stats.used_seats += len(license.active_users)
            
            # Expiring soon
            if license.days_until_expiry is not None:
                if license.days_until_expiry <= self._expiry_threshold_days:
                    stats.expiring_soon.append(license.name)
        
        return stats
    
    def add_alert_handler(self, handler: Callable) -> None:
        """Add alert handler."""
        self._alert_handlers.append(handler)


# Factory functions
def create_license_manager() -> LicenseManager:
    """Create license manager."""
    return LicenseManager()


def create_license(
    name: str,
    license_type: LicenseType = LicenseType.SUBSCRIPTION,
    **kwargs,
) -> License:
    """Create license."""
    return License(name=name, license_type=license_type, **kwargs)


__all__ = [
    # Exceptions
    "LicenseError",
    "LicenseExpired",
    "NoSeatsAvailable",
    # Enums
    "LicenseType",
    "LicenseStatus",
    "UsageEventType",
    # Data classes
    "UsageEvent",
    "LicenseKey",
    "License",
    "ComplianceIssue",
    "ComplianceReport",
    "LicenseStats",
    # Stores
    "LicenseStore",
    "InMemoryLicenseStore",
    "UsageStore",
    "InMemoryUsageStore",
    # Utilities
    "LicenseKeyGenerator",
    # Manager
    "LicenseManager",
    # Factory functions
    "create_license_manager",
    "create_license",
]



"""
Canary Deployment Manager.

Manages canary deployments for AI agents:
- Progressive traffic shifting
- Health monitoring
- Automatic rollback
- Deployment gates
"""

import uuid
import time
import logging
import random
import threading
from typing import Dict, Any, List
from collections import defaultdict

logger = logging.getLogger(__name__)


class CanaryDeploymentManager:
    """Manages canary deployments for AI agents."""
    
    def __init__(self):
        self.deployments: Dict[str, Dict[str, Any]] = {}
        self.deployment_history: List[Dict[str, Any]] = []
        self.health_checks: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
        self._lock = threading.Lock()
    
    def create_deployment(self,
                         name: str,
                         baseline_version: str,
                         canary_version: str,
                         traffic_steps: List[float] = None,
                         health_threshold: float = 0.95,
                         min_samples_per_step: int = 100) -> Dict[str, Any]:
        """Create a new canary deployment."""
        if traffic_steps is None:
            traffic_steps = [0.05, 0.25, 0.50, 1.0]
        
        deployment = {
            'id': str(uuid.uuid4()),
            'name': name,
            'baseline_version': baseline_version,
            'canary_version': canary_version,
            'traffic_steps': traffic_steps,
            'current_step': 0,
            'current_traffic': traffic_steps[0],
            'health_threshold': health_threshold,
            'min_samples_per_step': min_samples_per_step,
            'status': 'active',
            'created_at': time.time(),
            'metrics': {
                'baseline': {'success': 0, 'failure': 0, 'total_latency': 0},
                'canary': {'success': 0, 'failure': 0, 'total_latency': 0}
            }
        }
        
        with self._lock:
            self.deployments[name] = deployment
        
        logger.info("Created canary deployment '%s': %s -> %s",
                   name, baseline_version, canary_version)
        return deployment
    
    def route_request(self, deployment_name: str) -> str:
        """Route a request to baseline or canary."""
        deployment = self.deployments.get(deployment_name)
        if not deployment or deployment['status'] != 'active':
            return 'baseline'
        return 'canary' if random.random() < deployment['current_traffic'] else 'baseline'
    
    def record_result(self,
                     deployment_name: str,
                     version: str,
                     success: bool,
                     latency_ms: float = 0) -> Dict[str, Any]:
        """Record a request result."""
        deployment = self.deployments.get(deployment_name)
        if not deployment:
            return {'error': 'Deployment not found'}
        
        with self._lock:
            metrics = deployment['metrics'][version]
            if success:
                metrics['success'] += 1
            else:
                metrics['failure'] += 1
            metrics['total_latency'] += latency_ms
        
        return self._evaluate_deployment(deployment_name)
    
    def _evaluate_deployment(self, deployment_name: str) -> Dict[str, Any]:
        """Evaluate deployment health and decide next action."""
        deployment = self.deployments.get(deployment_name)
        if not deployment or deployment['status'] != 'active':
            return {'action': 'none', 'reason': 'deployment not active'}
        
        canary_metrics = deployment['metrics']['canary']
        canary_total = canary_metrics['success'] + canary_metrics['failure']
        
        if canary_total < deployment['min_samples_per_step']:
            return {
                'action': 'wait',
                'reason': f'need {deployment["min_samples_per_step"] - canary_total} more samples'
            }
        
        success_rate = canary_metrics['success'] / canary_total if canary_total > 0 else 0
        
        if success_rate < deployment['health_threshold']:
            self._rollback(deployment_name, f'success rate {success_rate:.2%} below threshold')
            return {
                'action': 'rollback',
                'reason': f'success rate {success_rate:.2%} < {deployment["health_threshold"]:.2%}'
            }
        
        current_step = deployment['current_step']
        traffic_steps = deployment['traffic_steps']
        
        if current_step < len(traffic_steps) - 1:
            deployment['current_step'] += 1
            deployment['current_traffic'] = traffic_steps[deployment['current_step']]
            deployment['metrics']['canary'] = {'success': 0, 'failure': 0, 'total_latency': 0}
            
            logger.info("Advancing canary '%s' to %.0f%% traffic",
                       deployment_name, deployment['current_traffic'] * 100)
            
            return {
                'action': 'advance',
                'new_traffic': deployment['current_traffic'],
                'step': deployment['current_step']
            }
        else:
            self._complete_deployment(deployment_name)
            return {'action': 'complete', 'reason': 'all steps passed'}
    
    def _rollback(self, deployment_name: str, reason: str):
        """Rollback a deployment."""
        deployment = self.deployments.get(deployment_name)
        if deployment:
            deployment['status'] = 'rolled_back'
            deployment['rollback_reason'] = reason
            deployment['completed_at'] = time.time()
            
            self.deployment_history.append({
                'deployment': deployment.copy(),
                'outcome': 'rolled_back',
                'timestamp': time.time()
            })
            logger.warning("Rolled back canary '%s': %s", deployment_name, reason)
    
    def _complete_deployment(self, deployment_name: str):
        """Complete a successful deployment."""
        deployment = self.deployments.get(deployment_name)
        if deployment:
            deployment['status'] = 'completed'
            deployment['completed_at'] = time.time()
            
            self.deployment_history.append({
                'deployment': deployment.copy(),
                'outcome': 'completed',
                'timestamp': time.time()
            })
            logger.info("Completed canary deployment '%s'", deployment_name)
    
    def get_status(self, deployment_name: str) -> Dict[str, Any]:
        """Get deployment status and metrics."""
        deployment = self.deployments.get(deployment_name)
        if not deployment:
            return {'error': 'Deployment not found'}
        
        baseline = deployment['metrics']['baseline']
        canary = deployment['metrics']['canary']
        
        baseline_total = baseline['success'] + baseline['failure']
        canary_total = canary['success'] + canary['failure']
        
        return {
            **deployment,
            'baseline_success_rate': baseline['success'] / baseline_total if baseline_total > 0 else 0,
            'canary_success_rate': canary['success'] / canary_total if canary_total > 0 else 0,
            'baseline_avg_latency': baseline['total_latency'] / baseline_total if baseline_total > 0 else 0,
            'canary_avg_latency': canary['total_latency'] / canary_total if canary_total > 0 else 0
        }
    
    def list_deployments(self, status: str = None) -> List[Dict[str, Any]]:
        """List all deployments, optionally filtered by status."""
        deployments = list(self.deployments.values())
        if status:
            deployments = [d for d in deployments if d['status'] == status]
        return deployments
    
    def abort_deployment(self, deployment_name: str, reason: str = 'manual abort'):
        """Manually abort a deployment."""
        self._rollback(deployment_name, reason)


__all__ = ['CanaryDeploymentManager']



"""
Agent Communication Module.

Provides multi-protocol agent-to-agent communication:

Protocols:
- STDIO: Standard input/output for local process communication (MCP-style)
- HTTP/HTTPS: RESTful communication for remote agents
- SSE: Server-Sent Events for real-time streaming responses
- MQTT: Message queue for IoT and distributed agent systems
- WebSocket: Bidirectional real-time communication

Features:
- RemoteAgentClient: Call remote agents over any protocol
- RemoteAgentServer: Expose agents as services (Flask, FastAPI, MQTT)
- AgentCommunicationManager: Central hub for all agent communications
- Message routing, broadcasting, and handoff support

Example:
    >>> from agenticaiframework.communication import (
    ...     AgentCommunicationManager,
    ...     RemoteAgentClient,
    ...     AgentEndpoint,
    ...     ProtocolType,
    ... )
    >>> 
    >>> # Using the manager (recommended)
    >>> manager = AgentCommunicationManager(agent_id="orchestrator")
    >>> manager.register_agent("analyzer", "https://analyzer.example.com/agent")
    >>> response = manager.send("analyzer", "Analyze this document")
    >>> 
    >>> # Using client directly
    >>> client = RemoteAgentClient()
    >>> client.register_endpoint(AgentEndpoint(
    ...     agent_id="summarizer",
    ...     protocol=ProtocolType.SSE,
    ...     host="summarizer.example.com"
    ... ))
    >>> for chunk in client.call_stream("summarizer", "Summarize article"):
    ...     print(chunk)
"""

from .protocols import (
    ProtocolType,
    ProtocolConfig,
    CommunicationProtocol,
    STDIOProtocol,
    HTTPProtocol,
    SSEProtocol,
    MQTTProtocol,
    WebSocketProtocol,
)
from .agent_channel import (
    AgentChannel,
    AgentMessage,
    MessageType,
)
from .remote_agent import (
    RemoteAgentClient,
    RemoteAgentServer,
    AgentEndpoint,
)
from .manager import AgentCommunicationManager


# Legacy CommunicationManager for backward compatibility
class CommunicationManager:
    """
    Legacy CommunicationManager for backward compatibility.
    
    For new code, use AgentCommunicationManager instead.
    """
    
    def __init__(self):
        from typing import Dict, Any, Callable
        self.protocols: Dict[str, Callable[[Any], Any]] = {}
    
    def _log(self, message: str):
        import time
        import logging as _logging
        _logging.getLogger(__name__).info("[CommunicationManager] %s", message)

    def register_protocol(self, name: str, handler_fn):
        self.protocols[name] = handler_fn
        self._log(f"Registered communication protocol '{name}'")

    def register_handler(self, handler_fn, name: str = None):
        """Alternative method for registering handlers"""
        protocol_name = name or f"handler_{len(self.protocols)}"
        self.register_protocol(protocol_name, handler_fn)

    def send(self, protocol: str, data):
        if protocol in self.protocols:
            try:
                return self.protocols[protocol](data)
            except (TypeError, ValueError, RuntimeError, ConnectionError, OSError) as e:
                self._log(f"Error sending data via '{protocol}': {e}")
                return None
        self._log(f"Protocol '{protocol}' not found")
        return None

    def list_protocols(self):
        return list(self.protocols.keys())

    def send_message(self, message, protocol: str = None):
        """Send a message using the first available protocol or specified protocol"""
        if protocol:
            return self.send(protocol, message)
        elif self.protocols:
            first_protocol = next(iter(self.protocols))
            return self.send(first_protocol, message)
        return None


__all__ = [
    # Legacy (backward compatibility)
    "CommunicationManager",
    # Protocol Types
    "ProtocolType",
    "ProtocolConfig",
    # Protocols
    "CommunicationProtocol",
    "STDIOProtocol",
    "HTTPProtocol",
    "SSEProtocol",
    "MQTTProtocol",
    "WebSocketProtocol",
    # Channel & Messages
    "AgentChannel",
    "AgentMessage",
    "MessageType",
    # Remote Agent
    "RemoteAgentClient",
    "RemoteAgentServer",
    "AgentEndpoint",
    # Manager
    "AgentCommunicationManager",
]



"""
Enterprise Graph Module.

Provides knowledge graph structures, entity relationships,
graph traversal, and graph-based reasoning for agents.

Example:
    # Create graph
    graph = KnowledgeGraph()
    
    # Add nodes and edges
    graph.add_node("Alice", type="person", age=30)
    graph.add_node("Bob", type="person", age=25)
    graph.add_edge("Alice", "Bob", relation="knows")
    
    # Query
    friends = graph.neighbors("Alice", relation="knows")
    path = graph.shortest_path("Alice", "Charlie")
"""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from collections import defaultdict, deque
from datetime import datetime
from functools import wraps
from enum import Enum
import logging
import json
import hashlib

logger = logging.getLogger(__name__)

T = TypeVar('T')
NodeId = str
EdgeId = str


class GraphError(Exception):
    """Graph operation error."""
    pass


class NodeNotFoundError(GraphError):
    """Node not found."""
    pass


class EdgeNotFoundError(GraphError):
    """Edge not found."""
    pass


class CycleDetectedError(GraphError):
    """Cycle detected in graph."""
    pass


class NodeType(str, Enum):
    """Common node types."""
    ENTITY = "entity"
    CONCEPT = "concept"
    DOCUMENT = "document"
    PERSON = "person"
    ORGANIZATION = "organization"
    LOCATION = "location"
    EVENT = "event"


class EdgeType(str, Enum):
    """Common edge/relation types."""
    RELATES_TO = "relates_to"
    IS_A = "is_a"
    HAS_A = "has_a"
    PART_OF = "part_of"
    MENTIONS = "mentions"
    SIMILAR_TO = "similar_to"
    DERIVED_FROM = "derived_from"


@dataclass
class Node:
    """A graph node."""
    id: NodeId
    type: str = "entity"
    properties: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.now)
    embedding: Optional[List[float]] = None
    
    def __hash__(self) -> int:
        return hash(self.id)
    
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Node):
            return self.id == other.id
        return self.id == other
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "type": self.type,
            "properties": self.properties,
            "created_at": self.created_at.isoformat(),
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'Node':
        return cls(
            id=data["id"],
            type=data.get("type", "entity"),
            properties=data.get("properties", {}),
        )


@dataclass
class Edge:
    """A graph edge."""
    id: EdgeId
    source: NodeId
    target: NodeId
    relation: str = "relates_to"
    weight: float = 1.0
    properties: Dict[str, Any] = field(default_factory=dict)
    directed: bool = True
    created_at: datetime = field(default_factory=datetime.now)
    
    def __hash__(self) -> int:
        return hash(self.id)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "source": self.source,
            "target": self.target,
            "relation": self.relation,
            "weight": self.weight,
            "properties": self.properties,
            "directed": self.directed,
        }


@dataclass
class Path:
    """A path through the graph."""
    nodes: List[Node]
    edges: List[Edge]
    
    @property
    def length(self) -> int:
        return len(self.edges)
    
    @property
    def weight(self) -> float:
        return sum(e.weight for e in self.edges)
    
    @property
    def node_ids(self) -> List[NodeId]:
        return [n.id for n in self.nodes]


@dataclass
class Subgraph:
    """A subgraph view."""
    nodes: Set[Node]
    edges: Set[Edge]
    
    @property
    def node_count(self) -> int:
        return len(self.nodes)
    
    @property
    def edge_count(self) -> int:
        return len(self.edges)


class Graph(ABC):
    """Abstract graph interface."""
    
    @abstractmethod
    def add_node(
        self,
        id: NodeId,
        type: str = "entity",
        **properties: Any,
    ) -> Node:
        """Add a node."""
        pass
    
    @abstractmethod
    def add_edge(
        self,
        source: NodeId,
        target: NodeId,
        relation: str = "relates_to",
        weight: float = 1.0,
        **properties: Any,
    ) -> Edge:
        """Add an edge."""
        pass
    
    @abstractmethod
    def get_node(self, id: NodeId) -> Optional[Node]:
        """Get a node by ID."""
        pass
    
    @abstractmethod
    def get_edge(self, source: NodeId, target: NodeId) -> Optional[Edge]:
        """Get an edge."""
        pass
    
    @abstractmethod
    def neighbors(
        self,
        id: NodeId,
        direction: str = "out",
        relation: Optional[str] = None,
    ) -> List[Node]:
        """Get neighboring nodes."""
        pass
    
    @abstractmethod
    def remove_node(self, id: NodeId) -> bool:
        """Remove a node."""
        pass
    
    @abstractmethod
    def remove_edge(self, source: NodeId, target: NodeId) -> bool:
        """Remove an edge."""
        pass


class KnowledgeGraph(Graph):
    """
    In-memory knowledge graph.
    """
    
    def __init__(self):
        self._nodes: Dict[NodeId, Node] = {}
        self._edges: Dict[EdgeId, Edge] = {}
        self._adjacency: Dict[NodeId, Dict[NodeId, EdgeId]] = defaultdict(dict)  # out edges
        self._reverse_adjacency: Dict[NodeId, Dict[NodeId, EdgeId]] = defaultdict(dict)  # in edges
    
    @property
    def node_count(self) -> int:
        return len(self._nodes)
    
    @property
    def edge_count(self) -> int:
        return len(self._edges)
    
    def add_node(
        self,
        id: NodeId,
        type: str = "entity",
        **properties: Any,
    ) -> Node:
        """Add or update a node."""
        if id in self._nodes:
            node = self._nodes[id]
            node.type = type
            node.properties.update(properties)
        else:
            node = Node(id=id, type=type, properties=properties)
            self._nodes[id] = node
        
        return node
    
    def add_edge(
        self,
        source: NodeId,
        target: NodeId,
        relation: str = "relates_to",
        weight: float = 1.0,
        directed: bool = True,
        **properties: Any,
    ) -> Edge:
        """Add an edge."""
        # Ensure nodes exist
        if source not in self._nodes:
            self.add_node(source)
        if target not in self._nodes:
            self.add_node(target)
        
        edge_id = f"{source}-{relation}-{target}"
        
        edge = Edge(
            id=edge_id,
            source=source,
            target=target,
            relation=relation,
            weight=weight,
            directed=directed,
            properties=properties,
        )
        
        self._edges[edge_id] = edge
        self._adjacency[source][target] = edge_id
        self._reverse_adjacency[target][source] = edge_id
        
        # For undirected edges, add reverse
        if not directed:
            rev_id = f"{target}-{relation}-{source}"
            rev_edge = Edge(
                id=rev_id,
                source=target,
                target=source,
                relation=relation,
                weight=weight,
                directed=directed,
                properties=properties,
            )
            self._edges[rev_id] = rev_edge
            self._adjacency[target][source] = rev_id
            self._reverse_adjacency[source][target] = rev_id
        
        return edge
    
    def get_node(self, id: NodeId) -> Optional[Node]:
        """Get a node."""
        return self._nodes.get(id)
    
    def get_edge(self, source: NodeId, target: NodeId) -> Optional[Edge]:
        """Get an edge."""
        edge_id = self._adjacency.get(source, {}).get(target)
        if edge_id:
            return self._edges.get(edge_id)
        return None
    
    def get_edges_between(
        self,
        source: NodeId,
        target: NodeId,
    ) -> List[Edge]:
        """Get all edges between two nodes."""
        edges = []
        
        # Check forward
        if source in self._adjacency and target in self._adjacency[source]:
            edges.append(self._edges[self._adjacency[source][target]])
        
        # Check reverse
        if target in self._adjacency and source in self._adjacency[target]:
            edges.append(self._edges[self._adjacency[target][source]])
        
        return edges
    
    def neighbors(
        self,
        id: NodeId,
        direction: str = "out",
        relation: Optional[str] = None,
    ) -> List[Node]:
        """Get neighboring nodes."""
        if id not in self._nodes:
            return []
        
        neighbors = []
        
        if direction in ("out", "both"):
            for target_id, edge_id in self._adjacency.get(id, {}).items():
                edge = self._edges[edge_id]
                if relation is None or edge.relation == relation:
                    neighbors.append(self._nodes[target_id])
        
        if direction in ("in", "both"):
            for source_id, edge_id in self._reverse_adjacency.get(id, {}).items():
                edge = self._edges[edge_id]
                if relation is None or edge.relation == relation:
                    if self._nodes[source_id] not in neighbors:
                        neighbors.append(self._nodes[source_id])
        
        return neighbors
    
    def edges_of(
        self,
        id: NodeId,
        direction: str = "out",
        relation: Optional[str] = None,
    ) -> List[Edge]:
        """Get edges connected to a node."""
        if id not in self._nodes:
            return []
        
        edges = []
        
        if direction in ("out", "both"):
            for edge_id in self._adjacency.get(id, {}).values():
                edge = self._edges[edge_id]
                if relation is None or edge.relation == relation:
                    edges.append(edge)
        
        if direction in ("in", "both"):
            for edge_id in self._reverse_adjacency.get(id, {}).values():
                edge = self._edges[edge_id]
                if relation is None or edge.relation == relation:
                    if edge not in edges:
                        edges.append(edge)
        
        return edges
    
    def remove_node(self, id: NodeId) -> bool:
        """Remove a node and its edges."""
        if id not in self._nodes:
            return False
        
        # Remove edges
        for target_id in list(self._adjacency.get(id, {}).keys()):
            self.remove_edge(id, target_id)
        
        for source_id in list(self._reverse_adjacency.get(id, {}).keys()):
            self.remove_edge(source_id, id)
        
        del self._nodes[id]
        return True
    
    def remove_edge(self, source: NodeId, target: NodeId) -> bool:
        """Remove an edge."""
        edge_id = self._adjacency.get(source, {}).get(target)
        
        if not edge_id:
            return False
        
        del self._edges[edge_id]
        del self._adjacency[source][target]
        del self._reverse_adjacency[target][source]
        
        return True
    
    def shortest_path(
        self,
        source: NodeId,
        target: NodeId,
        relation: Optional[str] = None,
    ) -> Optional[Path]:
        """Find shortest path using BFS."""
        if source not in self._nodes or target not in self._nodes:
            return None
        
        if source == target:
            return Path(nodes=[self._nodes[source]], edges=[])
        
        # BFS
        visited = {source}
        queue = deque([(source, [source], [])])
        
        while queue:
            current, node_path, edge_path = queue.popleft()
            
            for neighbor in self.neighbors(current, direction="out", relation=relation):
                if neighbor.id in visited:
                    continue
                
                edge = self.get_edge(current, neighbor.id)
                new_node_path = node_path + [neighbor.id]
                new_edge_path = edge_path + [edge]
                
                if neighbor.id == target:
                    return Path(
                        nodes=[self._nodes[n] for n in new_node_path],
                        edges=new_edge_path,
                    )
                
                visited.add(neighbor.id)
                queue.append((neighbor.id, new_node_path, new_edge_path))
        
        return None
    
    def all_paths(
        self,
        source: NodeId,
        target: NodeId,
        max_depth: int = 5,
    ) -> List[Path]:
        """Find all paths between nodes."""
        if source not in self._nodes or target not in self._nodes:
            return []
        
        paths = []
        
        def dfs(current: NodeId, path: List[NodeId], edges: List[Edge], depth: int):
            if depth > max_depth:
                return
            
            if current == target:
                paths.append(Path(
                    nodes=[self._nodes[n] for n in path],
                    edges=edges,
                ))
                return
            
            for neighbor in self.neighbors(current, direction="out"):
                if neighbor.id not in path:
                    edge = self.get_edge(current, neighbor.id)
                    dfs(
                        neighbor.id,
                        path + [neighbor.id],
                        edges + [edge],
                        depth + 1,
                    )
        
        dfs(source, [source], [], 0)
        return paths
    
    def subgraph(
        self,
        node_ids: Set[NodeId],
        include_edges: bool = True,
    ) -> Subgraph:
        """Extract a subgraph."""
        nodes = {self._nodes[id] for id in node_ids if id in self._nodes}
        edges = set()
        
        if include_edges:
            for edge in self._edges.values():
                if edge.source in node_ids and edge.target in node_ids:
                    edges.add(edge)
        
        return Subgraph(nodes=nodes, edges=edges)
    
    def neighborhood(
        self,
        id: NodeId,
        hops: int = 1,
        relation: Optional[str] = None,
    ) -> Subgraph:
        """Get k-hop neighborhood."""
        if id not in self._nodes:
            return Subgraph(nodes=set(), edges=set())
        
        node_ids = {id}
        frontier = {id}
        
        for _ in range(hops):
            new_frontier = set()
            for node_id in frontier:
                for neighbor in self.neighbors(node_id, direction="both", relation=relation):
                    if neighbor.id not in node_ids:
                        new_frontier.add(neighbor.id)
                        node_ids.add(neighbor.id)
            frontier = new_frontier
        
        return self.subgraph(node_ids)
    
    def find_nodes(
        self,
        type: Optional[str] = None,
        **properties: Any,
    ) -> List[Node]:
        """Find nodes by type and properties."""
        results = []
        
        for node in self._nodes.values():
            if type and node.type != type:
                continue
            
            match = True
            for key, value in properties.items():
                if node.properties.get(key) != value:
                    match = False
                    break
            
            if match:
                results.append(node)
        
        return results
    
    def find_edges(
        self,
        relation: Optional[str] = None,
        **properties: Any,
    ) -> List[Edge]:
        """Find edges by relation and properties."""
        results = []
        
        for edge in self._edges.values():
            if relation and edge.relation != relation:
                continue
            
            match = True
            for key, value in properties.items():
                if edge.properties.get(key) != value:
                    match = False
                    break
            
            if match:
                results.append(edge)
        
        return results
    
    def degree(self, id: NodeId, direction: str = "both") -> int:
        """Get node degree."""
        count = 0
        
        if direction in ("out", "both"):
            count += len(self._adjacency.get(id, {}))
        
        if direction in ("in", "both"):
            count += len(self._reverse_adjacency.get(id, {}))
        
        return count
    
    def pagerank(
        self,
        damping: float = 0.85,
        iterations: int = 100,
        tolerance: float = 1e-6,
    ) -> Dict[NodeId, float]:
        """Calculate PageRank scores."""
        n = len(self._nodes)
        if n == 0:
            return {}
        
        # Initialize
        scores = {id: 1.0 / n for id in self._nodes}
        
        for _ in range(iterations):
            new_scores = {}
            diff = 0
            
            for node_id in self._nodes:
                rank = (1 - damping) / n
                
                # Sum contributions from incoming edges
                for source_id in self._reverse_adjacency.get(node_id, {}):
                    out_degree = len(self._adjacency.get(source_id, {}))
                    if out_degree > 0:
                        rank += damping * scores[source_id] / out_degree
                
                new_scores[node_id] = rank
                diff += abs(rank - scores[node_id])
            
            scores = new_scores
            
            if diff < tolerance:
                break
        
        return scores
    
    def to_dict(self) -> Dict[str, Any]:
        """Export graph to dictionary."""
        return {
            "nodes": [n.to_dict() for n in self._nodes.values()],
            "edges": [e.to_dict() for e in self._edges.values()],
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'KnowledgeGraph':
        """Import graph from dictionary."""
        graph = cls()
        
        for node_data in data.get("nodes", []):
            graph.add_node(
                node_data["id"],
                type=node_data.get("type", "entity"),
                **node_data.get("properties", {}),
            )
        
        for edge_data in data.get("edges", []):
            graph.add_edge(
                edge_data["source"],
                edge_data["target"],
                relation=edge_data.get("relation", "relates_to"),
                weight=edge_data.get("weight", 1.0),
                **edge_data.get("properties", {}),
            )
        
        return graph
    
    def clear(self) -> None:
        """Clear the graph."""
        self._nodes.clear()
        self._edges.clear()
        self._adjacency.clear()
        self._reverse_adjacency.clear()


class EntityExtractor:
    """Extract entities and relations from text."""
    
    def __init__(
        self,
        llm_client: Optional[Any] = None,
        model: Optional[str] = None,
    ):
        self._client = llm_client
        self._model = model
    
    async def extract(
        self,
        text: str,
        entity_types: Optional[List[str]] = None,
        relation_types: Optional[List[str]] = None,
    ) -> Tuple[List[Node], List[Edge]]:
        """Extract entities and relations from text."""
        if not self._client:
            return [], []
        
        # Build prompt
        prompt = f"""Extract entities and relationships from the following text.

Text: {text}

Return JSON with:
- entities: list of {{id, type, properties}}
- relations: list of {{source, target, relation, properties}}

Entity types: {entity_types or ["person", "organization", "location", "concept"]}
Relation types: {relation_types or ["relates_to", "works_for", "located_in", "is_a"]}

JSON:"""
        
        response = await self._client.chat.completions.create(
            model=self._model,
            messages=[{"role": "user", "content": prompt}],
            temperature=0,
        )
        
        try:
            content = response.choices[0].message.content
            # Parse JSON
            import re
            match = re.search(r'\{.*\}', content, re.DOTALL)
            if match:
                data = json.loads(match.group())
                
                nodes = [
                    Node(
                        id=e["id"],
                        type=e.get("type", "entity"),
                        properties=e.get("properties", {}),
                    )
                    for e in data.get("entities", [])
                ]
                
                edges = [
                    Edge(
                        id=f"{r['source']}-{r['relation']}-{r['target']}",
                        source=r["source"],
                        target=r["target"],
                        relation=r.get("relation", "relates_to"),
                        properties=r.get("properties", {}),
                    )
                    for r in data.get("relations", [])
                ]
                
                return nodes, edges
        except Exception as e:
            logger.error(f"Failed to parse extraction: {e}")
        
        return [], []


def create_graph() -> KnowledgeGraph:
    """Factory function to create a knowledge graph."""
    return KnowledgeGraph()


__all__ = [
    # Exceptions
    "GraphError",
    "NodeNotFoundError",
    "EdgeNotFoundError",
    "CycleDetectedError",
    # Enums
    "NodeType",
    "EdgeType",
    # Data classes
    "Node",
    "Edge",
    "Path",
    "Subgraph",
    # Graph
    "Graph",
    "KnowledgeGraph",
    # Extraction
    "EntityExtractor",
    # Factory
    "create_graph",
    # Type aliases
    "NodeId",
    "EdgeId",
]



"""
Model Registry.

Pre-configured model definitions for common providers (2026 pricing).
"""

from typing import Dict
from .types import ModelTier, ModelCapability, ModelConfig


# Pre-configured model definitions for common providers (2026 pricing)
MODEL_REGISTRY: Dict[str, ModelConfig] = {
    # OpenAI Models
    'gpt-4o': ModelConfig(
        name='gpt-4o', tier=ModelTier.LLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.VISION,
                     ModelCapability.FUNCTION_CALLING, ModelCapability.STRUCTURED_OUTPUT],
        max_tokens=16384, context_window=128000, supports_streaming=True,
        supports_json_mode=True, cost_per_1k_input=0.0025, cost_per_1k_output=0.01,
        latency_ms_avg=800, provider='openai', version='2024-11'
    ),
    'gpt-4o-mini': ModelConfig(
        name='gpt-4o-mini', tier=ModelTier.SLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.FUNCTION_CALLING],
        max_tokens=16384, context_window=128000, supports_streaming=True,
        supports_json_mode=True, cost_per_1k_input=0.00015, cost_per_1k_output=0.0006,
        latency_ms_avg=400, provider='openai', version='2024-07'
    ),
    'o1': ModelConfig(
        name='o1', tier=ModelTier.RLM,
        capabilities=[ModelCapability.REASONING, ModelCapability.CHAIN_OF_THOUGHT,
                     ModelCapability.CODE_GENERATION],
        max_tokens=100000, context_window=200000, supports_streaming=True,
        cost_per_1k_input=0.015, cost_per_1k_output=0.06,
        latency_ms_avg=5000, provider='openai', version='2024-12'
    ),
    'o1-mini': ModelConfig(
        name='o1-mini', tier=ModelTier.RLM,
        capabilities=[ModelCapability.REASONING, ModelCapability.CHAIN_OF_THOUGHT],
        max_tokens=65536, context_window=128000, supports_streaming=True,
        cost_per_1k_input=0.003, cost_per_1k_output=0.012,
        latency_ms_avg=2000, provider='openai', version='2024-09'
    ),
    'o3-mini': ModelConfig(
        name='o3-mini', tier=ModelTier.RLM,
        capabilities=[ModelCapability.REASONING, ModelCapability.CHAIN_OF_THOUGHT,
                     ModelCapability.CODE_GENERATION],
        max_tokens=100000, context_window=200000, supports_streaming=True,
        cost_per_1k_input=0.0011, cost_per_1k_output=0.0044,
        latency_ms_avg=1500, provider='openai', version='2025-01'
    ),
    
    # Anthropic Models
    'claude-4-opus': ModelConfig(
        name='claude-4-opus', tier=ModelTier.LLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.VISION,
                     ModelCapability.REASONING, ModelCapability.CODE_GENERATION],
        max_tokens=8192, context_window=200000, supports_streaming=True,
        cost_per_1k_input=0.015, cost_per_1k_output=0.075,
        latency_ms_avg=2000, provider='anthropic', version='2025-02'
    ),
    'claude-4-sonnet': ModelConfig(
        name='claude-4-sonnet', tier=ModelTier.LLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.VISION,
                     ModelCapability.FUNCTION_CALLING],
        max_tokens=8192, context_window=200000, supports_streaming=True,
        cost_per_1k_input=0.003, cost_per_1k_output=0.015,
        latency_ms_avg=1000, provider='anthropic', version='2025-02'
    ),
    'claude-3.5-haiku': ModelConfig(
        name='claude-3.5-haiku', tier=ModelTier.SLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.FUNCTION_CALLING],
        max_tokens=8192, context_window=200000, supports_streaming=True,
        cost_per_1k_input=0.0008, cost_per_1k_output=0.004,
        latency_ms_avg=300, provider='anthropic', version='2024-10'
    ),
    
    # Google Models
    'gemini-2.0-flash': ModelConfig(
        name='gemini-2.0-flash', tier=ModelTier.LLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.VISION,
                     ModelCapability.AUDIO, ModelCapability.FUNCTION_CALLING],
        max_tokens=8192, context_window=1000000, supports_streaming=True,
        supports_json_mode=True, cost_per_1k_input=0.0001, cost_per_1k_output=0.0004,
        latency_ms_avg=500, provider='google', version='2024-12'
    ),
    'gemini-2.0-flash-thinking': ModelConfig(
        name='gemini-2.0-flash-thinking', tier=ModelTier.RLM,
        capabilities=[ModelCapability.REASONING, ModelCapability.CHAIN_OF_THOUGHT],
        max_tokens=8192, context_window=1000000, supports_streaming=True,
        cost_per_1k_input=0.0, cost_per_1k_output=0.0,  # Free during preview
        latency_ms_avg=3000, provider='google', version='2024-12'
    ),
    'gemini-1.5-pro': ModelConfig(
        name='gemini-1.5-pro', tier=ModelTier.LLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.VISION,
                     ModelCapability.LONG_CONTEXT],
        max_tokens=8192, context_window=2000000, supports_streaming=True,
        cost_per_1k_input=0.00125, cost_per_1k_output=0.005,
        latency_ms_avg=1500, provider='google', version='2024-05'
    ),
    
    # Small Language Models (SLM)
    'phi-4': ModelConfig(
        name='phi-4', tier=ModelTier.SLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.REASONING],
        max_tokens=16384, context_window=16384, supports_streaming=True,
        cost_per_1k_input=0.0001, cost_per_1k_output=0.0004,
        latency_ms_avg=200, provider='microsoft', version='2024-12'
    ),
    'llama-3.3-70b': ModelConfig(
        name='llama-3.3-70b', tier=ModelTier.MLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.CODE_GENERATION],
        max_tokens=8192, context_window=128000, supports_streaming=True,
        cost_per_1k_input=0.0008, cost_per_1k_output=0.0008,
        latency_ms_avg=600, provider='meta', version='2024-12'
    ),
    'mistral-small': ModelConfig(
        name='mistral-small', tier=ModelTier.SLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.FUNCTION_CALLING],
        max_tokens=8192, context_window=32000, supports_streaming=True,
        cost_per_1k_input=0.0002, cost_per_1k_output=0.0006,
        latency_ms_avg=250, provider='mistral', version='2024-09'
    ),
    'qwen-2.5-72b': ModelConfig(
        name='qwen-2.5-72b', tier=ModelTier.MLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.CODE_GENERATION,
                     ModelCapability.REASONING],
        max_tokens=8192, context_window=131072, supports_streaming=True,
        cost_per_1k_input=0.0004, cost_per_1k_output=0.0004,
        latency_ms_avg=500, provider='alibaba', version='2024-09'
    ),
    
    # DeepSeek Reasoning Models
    'deepseek-r1': ModelConfig(
        name='deepseek-r1', tier=ModelTier.RLM,
        capabilities=[ModelCapability.REASONING, ModelCapability.CHAIN_OF_THOUGHT,
                     ModelCapability.CODE_GENERATION],
        max_tokens=8192, context_window=64000, supports_streaming=True,
        cost_per_1k_input=0.00055, cost_per_1k_output=0.00219,
        latency_ms_avg=2500, provider='deepseek', version='2025-01'
    ),
    'deepseek-v3': ModelConfig(
        name='deepseek-v3', tier=ModelTier.LLM,
        capabilities=[ModelCapability.TEXT_GENERATION, ModelCapability.CODE_GENERATION],
        max_tokens=8192, context_window=64000, supports_streaming=True,
        cost_per_1k_input=0.00027, cost_per_1k_output=0.0011,
        latency_ms_avg=800, provider='deepseek', version='2024-12'
    ),
}


__all__ = ['MODEL_REGISTRY']
